# Código para Autómata Finito [Determinístic]
from maquinas.exceptions import *
from maquinas.regular.dfa import *
from maquinas.contextfree.CFGParser import CFGParser
import maquinas.parser.earley_parser as earley_parser
from ordered_set import OrderedSet
from IPython.core.display import display, HTML
import random
import itertools
import copy
import re

parser=CFGParser()

def _flat(vals,ini=[]):
    if isinstance(vals,list):
        ini1=_flat(vals[0])
        ini2=_flat(vals[1])
        return ini1+ini2
    else:
        return [vals]

def _flat_tree(tree,acc=[]):
    if len(tree)==2 and tree[-1][-1]==";":
        acc=_flat_tree(tree[0])
        acc.append(tree[1])
        return acc
    return [tree]

## TODO Allow more constructive way to produce grammar
## TODO Allow to pass a tokeniser to grammar
class ContextFreeGrammar:
    """Class for Context Free Grammar

    :param er_string: String with grammar 
    :param S: Innitial symbol
    """
    def __init__(self, er_string, S=None):
        self.original = er_string
        self.Sigmariginal=er_string
        self.ast=parser.parse(er_string,rulne_name="start")
        if len(self.ast)==2:
            self.ast=_flat_tree(self.ast[0])+[self.ast[1]]
        else:
            self.ast=_flat_tree(self.ast[0])+[self.ast[1],self.ast[2]]
        self.V=set()
        self.sigma=set()
        self.P={}
        replacements={'epsilon':'ε'}
        for p in self.ast:
            head = p[0]
            if head not in self.sigma:
                self.V.add(head.replace('"',''))
        for p in self.ast:
            body=_flat(p[2])
            body=[replacements.get(p.replace('"',''),p.replace('"','')) for p in body]
            try:
                self.P[p[0]].append(body)
            except KeyError:
                self.P[p[0]]=[body]
            for val in body:
                if not val in self.V:
                    if not val == "ε":
                        self.sigma.add(val.replace('"',''))
        if S:
            self.S=S
        else:
            self.S=self.ast[0][0]
        tokens=[t for t in self.V]+[t for t in self.sigma]
        tokens.sort(key=len,reverse=True)
        self.re_tokens=re.compile("([{}])".format("|".join(tokens)))

    def parse(self,string):
        """Parses a string with grammar

        :param string: String to be parsed

        :returns: Tuple with Roots of trres, chart parser and forest
        """
        tokens=self.re_tokens.findall(string)
        if not len("".join(tokens)) == len(string):
            return False,[],{}
        chart=earley_parser.parse(tokens,self.P,self.S,self.sigma)
        forest=earley_parser.extract_forest(self.S,chart,self.sigma,tokens)
        roots=earley_parser.verify_chart(chart,tokens,self.S)
        return roots,chart,forest

    def accepts(self,string):
        """Parses a string with grammar and determines if accepts it

        :param string: String to be parsed

        :returns: Tuple with Roots of trres, chart parser and forest
        """
        tokens=self.re_tokens.findall(string)
        if not len("".join(tokens)) == len(string):
            return False,[],{}
        chart=earley_parser.parse(tokens,self.P,self.S,self.sigma)
        roots=earley_parser.verify_chart(chart,tokens,self.S)
        return len(roots)>0

    def _reduce_tree(self,tree):
        if len(tree)==1:
            return tree
        head,children=tree
        subtree=[]
        for child in children:
            child_=self._reduce_tree(child)
            if child_[0]['partial']:
                for grandchild in child_[1]:
                    subtree.append(grandchild)
            else:
                subtree.append(child_)
        return (head,subtree)

    ## TODO: Adds strategis for derivation
    def derivation(self,tree):
        """Iterator over derivation of a tree

        :param tree: Tree to retrieve derivation form

        :returns: Tuple with Roots of trres, chart parser and forest
        """
        tree=self._reduce_tree(tree)
        prev=None
        for i,nodes in enumerate(self.delta_analyse_tree(tree)):
            if i==0:
                string=[("","") for _ in range(nodes[0][1],nodes[0][2])]
            for node in nodes:
                ini,fin=node[1],node[2]
                if ini != fin:
                    pre,label=string[ini]
                    string[ini]=(pre,node[0])
                else:
                    if ini==len(string):
                        string.append(("",""))
                    pre,label=string[ini]
                    string[ini]=(node[0],label)
            yield [l[0]+l[1] for l in string] 

    ## TODO: Adds formas for derivation (e.g., Latex)
    def print_derivations(self,string,ini=0,fin=None):
        """Print derivations over trees 

        :param string: String to be parsed
        :param ini: Initial tree to start printing from
        :param final: Final tree to print
        """
        roots,chart,forest=self.parse(string)
        if roots:
            trees=self.extract_trees(forest)
            if fin:
                trees=trees[ini:fin]
            else:
                trees=trees[ini:]
            for i,tree in enumerate(trees):
                print("Tree #",ini+i)
                for j,step in enumerate(self.derivation(tree)):
                    if not j:
                        print("".join(step),end="")
                    else:
                        print(" ⇒ ","".join(step),end="\n ")
                print()

    ## TODO adds replacement of symbols
    def summary(self):
        """Generates summary of grammar
        """
        info= [
         "No terminal : "+", ".join(self.V),
         "Terminals   : "+", ".join(self.sigma),
         "Start       : "+", ".join(self.S),
         "Productions :\n"+"\n".join(f" {alpha} → {' | '.join(''.join(beta) for beta in betas)}" for alpha,betas in self.P.items())]
        return "\n".join(info) 

    def print_summary(self):
        """Prints summary of grammar
        """
        return print(self.summary()) 

    def _delta_analyse_tree(self,tree):
        if len(tree)==2:
            _,children=tree
            children_=[t[0]['label'] for t in children]
            yield children_
            for child in children:
                yield from self._delta_analyse_tree(child)

    def delta_analyse_tree(self,tree):
        """Iteraror over analysis of a tree 

        :param tree: Tree
        """
        tree=self._reduce_tree(tree)
        current,children=tree
        yield [current['label']]
        children_=[t[0]['label'] for t in children]
        yield children_
        for child in children:
            yield from self._delta_analyse_tree(child)

    ## TODO associate string to chart
    def print_chart(self,string,chart,pointers=False):
        """Prints chart parser

        :param string: Original string fom parser
        :param chart: Chart parser
        :param pointers: Include pointers (default is no)
        """
        earley_parser.print_chart(string,chart,pointers=pointers)

    ## TODO associate string to chart
    def chart2table(self,string,chart,pointers=False):
        """Creates HTML for chart parser

        :param string: Original string fom parser
        :param chart: Chart parser
        :param pointers: Include pointers (default is no)

        :returns: Display object for IPython"""
        earley_parser.chart2table(string,chart,pointers=pointers)

    def graph_forest(self,forest,**params):
        """Graph the forest structure

        :param forest: Forest structure
        :param params: Options to graph the tree

        :returns: Graphviz object"""
        return earley_parser.graph_forest(forest,**params)

    def extract_forest(self,chart):
        """Extract forest from chart object

        :param chart: Chart parser

        :returns: Forest data structure"""
        return earley_parser.extract_forest(self.S,chart,self.sigma)

    def extract_trees(self,forest,max_depth=None,max_ancesters=3):
        """Extract trees from forest structure

        :param forest: Forest data structure
        :param max_depth: Maximum depth to explore (default is None)
        :param max_ancester: Maximum number of the same ancester (default is 3)
        :returns: Trees data structure"""
        return earley_parser.extract_trees(forest,self.sigma,max_depth=max_depth,max_ancesters=max_ancesters)

    def graph_tree(self,tree,**params):
        """Graphs a tree

        :param tree: Tree data structure
        :param params: Options to graph the tree
        :returns: Graphviz objec"""
        return earley_parser.graph_tree(tree,**params)

    def graph_trees(self,trees,**params):
        """Graphs trees

        :param trees: Tree data structures
        :param params: Options to graph the tree
        :returns: Graphviz objec"""
        return earley_parser.graph_trees(trees,**params)

    def save_trees_img(self,trees,filename,format="svg",dpi="60.0",**params):
        """Saves trees into an image

        :param trees: List of tree data structures
        :param filename: Filename for the image file
        :param format: Image format (defualt svg)
        :param dpi: DPIs of fomat (default "60.0")
        :param params: Options to graph the tree"""
        dot=self.graph_trees(trees,**params)
        dot.render(filename,format=format,cleanup=True)

    def save_tree_img(self,tree,filename,format="svg",dpi="60.0",**params):
        """Saves a tree into an image

        :param tree: Tree data structures
        :param filename: Filename for the image file
        :param format: Image format (defualt svg)
        :param dpi: DPIs of fomat (default "60.0")
        :param params: Options to graph the tree"""
        dot=self.graph_tree(tree,**params)
        dot.render(filename,format=format,cleanup=True)

    def __str__(self):
        return self.original



import maquinas.regular
import maquinas.contextfree
import maquinas.contextsensitive
import maquinas.recursivelyenumerable
import maquinas.parser

import maquinas.exceptions
from .simulation import Simulation

__version__ = "0.1.5.18"



# Base class for Finite Machines
from maquinas.exceptions import *
from ordered_set import OrderedSet
import tempfile
import os
from PIL import Image
from IPython.core.display import display, HTML, Markdown, clear_output

class FiniteAutomaton():
    """Common class for finite automaton
    
    :param Q: Ordered set of states (default is empty).
    :param sigma: Ordered set of symbols (default is empty).
    :param q_0: Initial state (default None).
    :param A: Set of acceptor states (default is empty).
    :param delta: List of transitions with the form tupla of tupla of q_i and a, and q_f (default is empty).
    :param force: If True and states or symbols do not exists create them (default is False).
    :type force: bool
    :param epsilon: The Finite Automaton includes epsilon (default is False)
    :type epsilon: bool
    """

    def __init__(self, Q=[], sigma=[], q_0=None, A=[], delta=[], force=False, epsilon=False):
        self.sigma=OrderedSet()
        if epsilon:
            self.sigma.add('ε')
        self.sigma.update(sigma)
        self.Q=OrderedSet(Q)
        self.set_initial_state(q_0,force=force)
        self.set_aceptors(A,force=force)
        self.ttable={}
        for (q_i,a),q_f in delta:
            self.add_transition(q_i,a,q_f,force=force)

    def __getitem__(self,key):
        q,a=key
        return self.get_transition(q,a)

    def _nstate(self,q):
        """ Gets index of state or states given a state or states

        :param q: State or states
        :returns: index of state q or indexes of states q """
        if isinstance(q,set) or isinstance(q,list):
            return set(self.Q.index(n) for n in q)
        if q is None:
            return set()
        else:
            return self.Q.index(q)

    def _nsymbol(self,a):
        """ Gets index of symbol given a symbol

        :param a: Symbol a
        :returns: index of symbol a"""
        return self.sigma.index(a)

    def _state(self,nq):
        """ Gets state or states given an index or indexes

        :param nq: Index of state
        :returns: State or states"""
        if isinstance(nq,set):
            return set(self.Q.items[n] for n in nq)
        else:
            return self.Q.items[nq]

    def _symbol(self,na):
        """ Gets symbol given an index of a symbol

        :param na: Index of symbol
        :returns: Symbol"""
        return self.sigma.items[na]

    def states(self):
        """ Gets states

        :returns: States of machine
        :rtype: list"""
        return list(self.Q)

    def symbols(self):
        """ Gets symbols

        :returns: Symbols of machine
        :rtype: list"""
        return list(self.sigma)

    def _transition(self,nq_i,na,nq_f):
        """ Gets transition triplet given the index of states and symbol

        :param nq_i: Index of source state
        :param na: Index of symbol
        :param nq_f: Index of destination state
        :returns: Transition tuple"""
        return (self._state(nq_i),self._symbol(na)),self._state(nq_f)

    def __setitem__(self,key,value):
        q,a=key
        return self.add_transition(q,a,value)

    def _get_transition(self,nq,na):
        """ Gets the index or indexes of the destintion states for a index state and symbol

        :param nq: Index of source state
        :param na: Index of symbol
        :returns: Index of destination state or states"""
        if isinstance(nq,set) or isinstance(nq,list):
                new=set()
                for nq_ in nq:
                    try:
                        new=new|self.ttable[nq_][na]
                    except KeyError:
                        pass
                return new
        else:
            try:
                return self.ttable[nq][na]
            except KeyError:
                raise DoesNotExistsTransition(nq,na)

    def get_transition(self,q,a):
        """ Gets the destintion state or states for state and symbol

        :param nq: Source state
        :param na: Symbol
        :returns: Destination state or states"""
        res=self._get_transition(self._nstate(q),self._nsymbol(a))
        if isinstance(res,set) or isinstance(res,list):
            return set([self._state(q) for q in res])
        else:
            return set([self._state(res)])


    def items(self):
        """ Iterator over the transitions

        :returns: Yeilds a tuple transition"""
        for nq_i,t_ in self.ttable.items():
            for na,nq_f in t_.items():
                yield self._transition(nq_i,na,nq_f)

    def summary(self,symbols={},states={}):
        """ Produces a summary of the AF

        :param symbols: Replacements of the symbols to print
        :param states: Replacements of the states to print
        :returns: List with summary"""
        info= [
         "States  : "+", ".join([symbols.get(q,q) for q in self.states()]),
         "Sigma   : "+", ".join([symbols.get(a,a) for a in self.symbols()]),
         "Initial : "+symbols.get(self.q_0,self.q_0),
         "Aceptors: "+", ".join([symbols.get(q,q) for q in self.A]),
         "Transitions:\n"+"\n".join(f" {states.get(q_i,q_i)},{symbols.get(a,a)} → {tuple(states.get(q_f,q_f) for q_f in q_fs)}" for (q_i,a),q_fs in self.items())]
        return "\n".join(info)

    def print_summary(self,symbols={},states={},**args):
        """ Print a summary of the AF

        :param symbols: Replacements of the symbols to print
        :param states: Replacements of the states to print
        :param args: Parameters for the print
        :returns: None"""
        print(self.summary(symbols=symbols,states=states),**args)

    def print_transitions(self,w,symbols={},states={},**args):
        """ Print a transition for the string w

        :param w: Replacements of the symbols to print
        :param states: Replacements of the states to print
        :param args: Parameters for the print
        :returns: None"""
        for q,a,w_ in self.delta_stepwise(w):
            if a:
                print(f"{symbols.get(a,a)} → {states.get(q,q)}", end=",\n ",**args)
            else:
                print(f"{states.get(q,q)}",end=",\n ",**args)

    def add_transition(self,q_i,a,q_f,force=False):
        """ Adds a transition

        :param q_i: Source state
        :param a: Symbol
        :param q_f: Destination state
        :returns: None"""
        if a=='epsilon':
            a='ε'
        try:
            na = self._nsymbol(a)
        except KeyError:
            if force:
                na=self.add_symbol(a)
            else:
                raise DoesNotExistsSymbol(a)
        try:
            nq_i = self._nstate(q_i)
        except KeyError:
            if force:
                nq_i=self.add_state(q_i)
            else:
                raise DoesNotExistsState(q_i)

        if  nq_i and nq_i in self.ttable and na in self.ttable[nq_i]:
            raise AlreadyExistsTransition(q_i,a,self)
        try:
            nq_f = self._nstate(q_f)
        except KeyError:
            if force:
                nq_f=self.add_state(q_f)
            else:
                raise DoesNotExistsState(q_f)
        try:
            self.ttable[nq_i][na]=nq_f
        except KeyError:
            self.ttable[nq_i]={}
            self.ttable[nq_i][na]=nq_f

    def add_state(self,q,initial=False):
        """ Adds a state

        :param q: State or states
        :param initial: Set state as a initial
        :returns: Indixes of state or states"""
        if initial:
            self.q_0=q
        if isinstance(q,(set,list)):
            return set(self.Q.add(q_) for q_ in q)
        else:
            if q in self.Q:
                raise AlreadyExistsState(q)
            return self.Q.add(q)

    def add_next_state(self,initial=False):
        """ Adds a state with a infered name based on the number of states q_max. If the name state is already defined it looks the following integer available.

        :param q: State or states
        :param initial: Set state as a initial
        :returns: Next state generated and integer"""
        max_ix=len(self.Q)
        while f"q_{max_ix}" in self.Q:
            max_ix+=1
        q=f"q_{max_ix}"
        self.Q.add(q)
        if initial:
            self.q_0=q
        return q,max_ix

    def add_symbol(self,a):
        """ Adds a symbol

        :param a: Symbol
        :returns: Indixes of symbol"""
        if a == "epsilon":
            a="ε"
        if a in self.sigma:
            raise AlreadyExistsSymbol(a)
        return self.sigma.add(a)

    def set_initial_state(self,q,force=False):
        """ Sets an initial state

        :param q: State
        :param force: If not defined it creates it (default is False)
        :returns: None"""
        if q is None:
            self.q_0=None
            return None
        if not q in self.Q:
            if force:
                self.add_state(q)
            else:
                raise DoesNotExistsState(q)
        self.q_0=q

    def get_initial_state(self):
        """ Gets an initial state

        :returns: State"""
        return self.q_0

    def set_aceptors(self,A,force=False):
        """ Sets aceptors states

        :param A: States
        :param force: If not defined it creates it (default is False)
        :returns: None"""
        if force:
            self.add_state(A)
        self.A=set(A)

    def get_aceptors(self):
        """ Gets aceptors states

        :returns: States"""
        return self.A

    def autorename(self,start=0,avoid=[]):
        """ Autorenames the states with a patter of q_n, where n is a consicutive integer

        :param start: Staring numbering (default is 0)
        :param avoid: List of labelings to avoid (default is empty)
        :returns: None"""
        replacements=[(q,f'q_{start+ix}') for q,ix in self.Q.map.items()]
        ix=len(self.Q)+start
        new_A=set()
        while len(replacements)>0:
            old,new=replacements.pop(0)
            if new == old:
                if old in self.A:
                    new_A.add(new)
                if old == self.q_0:
                    self.q_0=new
                continue
            elif new in self.Q or new in avoid:
                replacements.append((old,f'q_{ix}'))
                ix+=1
            else:
                self.replace_state(old,new)
                if old in self.A:
                    new_A.add(new)
                if old == self.q_0:
                    self.q_0=new
        self.A=new_A

    def remove_states(self,states):
        """ Remove states 

        :param states: List of states to remove
        :returns: None"""
        Q_=OrderedSet()
        ttable_={}
        for q in self.Q:
            if not q in states:
                Q_.add(q)

        ttable_={}
        for nq_i,tt in self.ttable.items():
            q_i=self.Q.items[nq_i]
            if q_i in states:
                continue
            nq_i_=Q_.map[q_i]
            for na,nqs in tt.items():
                if isinstance(nqs,set):
                    qs=[self.Q.items[q] for q in nqs ]
                    res=set([Q_.index(q) for q in qs if not q in states])
                else:
                    q=self.Q.items[nqs]
                    if q in states:
                        continue
                    res=Q_.index(q)
                try:
                    ttable_[nq_i_][na]=res
                except KeyError:
                    ttable_[nq_i_]={}
                    ttable_[nq_i_][na]=res
        self.Q=Q_
        self.ttable=ttable_

    def remove_sink_states(self):
        """ Remove states that do no go to any place

        :returns: None"""
        Q_=OrderedSet()
        idx_remove=set()
        for nq,q in enumerate(self.Q):
            if not q in self.A:
                if not nq in self.ttable:
                    idx_remove.add(nq)
                    continue
            destination=set()
            # If state is loopy
            for na,a in enumerate(self.sigma):
                try:
                    res=self.ttable[nq][na]
                except KeyError:
                    continue
                if isinstance(res,set):
                    destination=destination | res
                else:
                    destination.add(res)
            if len(destination)==1 and nq in destination and q not in self.A:
                idx_remove.add(nq)
                continue
            Q_.add(q)

        ttable_={}
        for nq_i_,q_i in enumerate(Q_):
            nq_i=self.Q.map[q_i]
            for na,a in enumerate(self.sigma):
                try:
                    res=self.ttable[nq_i][na]
                except KeyError:
                    continue
                if isinstance(res,set):
                    qs=[self.Q.items[nq] for nq in res if not nq in idx_remove]
                    res=set([Q_.index(q) for q in qs])
                else:
                    if res in idx_remove:
                        continue
                    q=self.Q.items[res]
                    res=Q_.index(q)
                try:
                    ttable_[nq_i_][na]=res
                except KeyError:
                    ttable_[nq_i_]={}
                    ttable_[nq_i_][na]=res
        self.Q=Q_
        self.ttable=ttable_

    def _status(self,status,states={},symbols={}):
        return status

    def step(self,q,a):
        return self.delta(q,a)

    def stepStatus(self,status):
        """ Gives a step and calculates new status for Simulation

        :param Status: Status
        :returns: None"""
        if status.state is None:
            q_c=self.q_0
        else:
            q_c=status.state

        a=status.get_symbol_tape()
        q_c=self.step(q_c,a)
        status.position+=1
        status.step+=1
        status.state=q_c

    def accepts(self,w):
        """ Checks if string is accepted

        :param w: String
        :returns: None"""
        return self.acceptor(self.delta_extended(None,w))

    def acceptor(self,q):
        """ Checks if state is an acceptor state

        :param q: State or states
        :returns: None"""
        if isinstance(q,set):
            if bool(q.intersection(self.A)):
                return True
        else:
            if q in self.A:
                return True
        return False

    def replace_symbol(self,old,new):
        """ Replace a symbol

        :param old: Symbol to be replaced
        :param new: Symbol to replace with
        :returns: None"""
        ix=self.sigma.index(old)
        del self.sigma.map[old]
        self.sigma.map[new]=ix
        self.sigma.items[ix]=new

    def replace_state(self,old,new):
        """ Replace a state

        :param old: State to be replaced
        :param new: State to replace with
        :returns: None"""
        if new in self.Q:
            raise AlreadyExistsState(new)
        ix=self.Q.index(old)
        del self.Q.map[old]
        self.Q.map[new]=ix
        self.Q.items[ix]=new
        if old == self.q_0:
            self.q_0=new

    def add_error_state(self,e_label="q_E"):
        """ Adds a error state (sink state)

        :param e_label: Label for the error state
        :returns: None"""
        empty_cell=False
        for nq,q in enumerate(self.Q):
            for na,a in enumerate(self.sigma):
                if not nq in self.ttable:
                    empty_cell=True
                    break
                elif not na in self.ttable[nq] or len(self.ttable[nq][na])==0:
                    empty_cell=True
                    break
            if empty_cell:
                break
        if empty_cell:
            ne=self.add_state(e_label)
            ne=self.Q.index(e_label)
            for nq,q in enumerate(self.Q):
                for na,a in enumerate(self.sigma):
                    if not nq in self.ttable:
                        self.ttable[nq]={}
                        self.ttable[nq][na]=set([ne])
                    elif not na in self.ttable[nq] or len(self.ttable[nq][na])==0 :
                        self.ttable[nq][na]=set([ne])

    def reachable_states(self):
        """ Calculate the set of states which are reacheble

        :returns: List of reachable states"""
        # https://en.wikipedia.org/wiki/DFA_minimization#Unreachable_states
        reachable=set([self.q_0])
        new=set([self.q_0])
        while len(new)>0:
            temp=set()
            for q in new:
                for a in self.sigma:
                    try:
                        states=self.get_transition(q,a)
                        if isinstance(states,set):
                            temp.update(states)
                        else:
                            temp.add(states)
                    except DoesNotExistsTransition:
                        pass
            new=temp.difference(reachable)
            reachable=reachable.union(new)
        return reachable

    def unreachable_states(self):
        """ Calculate the set of states which are unreacheble

        :returns: List of unreachable states"""
        return set(self.Q.difference(self.reachable_states()))

    def remove_unreachable(self):
        """ Removes the set of states which are unreacheble

        :returns: List of unreachable states"""
        self.remove_states(self.unreachable_states())

    def save_file(self,filename="machine.txt",order_Q=None,order_sigma=None):
        """Saves a file
        
        :param filename: Name of filename (default is machine.txt)
        :param order_Q: Order to print states
        :param order_sigma: Order to print vocabulary

        :returns: None
        """
        largest_sigma=max([len(a) for a in self.sigma])
        largest_q=max([len(q) for q in self.Q])
        largest_q+=3
        largest_cell=0
        for (q_i,a),q_f in self.items():
            largest_cell=max(largest_cell,len(",".join(q_f)))
        largest_cell+=2
        if not order_Q:
            order_Q=list(self.Q)
        if not order_sigma:
            order_sigma=list(self.sigma)
        with open(filename,'w') as f:
            print(" "*(largest_q+2),end="",file=f)
            print("|",end="",file=f)
            h="|".join("{: ^{width}}".format(a,width=largest_cell) for a in self.sigma)
            print(h,end="",file=f)
            print("|",file=f)
            for q_i in order_Q:
                q_l=q_i
                if q_i == self.q_0:
                    q_l=f"->{q_i}"
                if q_i in self.A:
                    q_l=f"{q_l}]"
                if q_i == self.q_0:
                    print(" {: <{width}}|".format(q_l,width=largest_q+1),end="",file=f)
                else:
                    print("   {: <{width}}|".format(q_l,width=largest_q-1),end="",file=f)
                line=[]
                for a in order_sigma:
                    try:
                        line.append(",".join(self[(q_i,a)]))
                    except DoesNotExistsTransition:
                        line.append("")
                r="|".join("{: ^{width}}".format(c,width=largest_cell) for c in line)
                print(r,end="",file=f)
                print("|",file=f)

    def save_img(self,filename,q_c=set(),a_c=set(),q_prev=set(),symbols={},states={},format='svg',dpi="60.0",string=None):
        """ Saves machine as an image

        :param filename: Filename of image
        :param q_c: Set of current states to be highlited (default is empty)
        :param a_c: Set of current symbols to be highlited (default is empty)
        :param q_prev: Set of previos states to be highlited (default is empty)
        :param symbols: Replacements of the symbols to show (default is empty)
        :param states: Replacements of the states to show (default is empty)
        :param format: Format of image (default is svg)
        :param dpi: Resolution of image (default is "60.0")
        :param string: Label of string being analysed (default is None)
        :returns: None"""
        dot=self.graph(q_c=q_c,a_c=a_c,q_prev=q_prev,symbols=symbols,states=states,format=format,dpi=dpi,string=string)
        dot.render(filename,format=format,cleanup=True)

    def save_gif(self,w,filename="maquina.gif",symbols={},states={},dpi="90",show=False,loop=0,duration=500):
        """ Saves an animation of  machine

        :param w: String to analysed during animation
        :param filename: Name of gif (default is "maquina.gif")
        :param symbols: Replacements of the symbols to show (default is empty)
        :param states: Replacements of the states to show (default is empty)
        :param dpi: Resolution of image (default is "90.0")
        :param show: In interactive mode return gif
        :param loop: Number of loops in annimation, cero is forever (default is 0)
        :param duration: Duration in msegs among steps (default is 500)
        :returns: None or HTML for Ipython"""
        dirpath = tempfile.mkdtemp()
        i=0
        images=[]
        q_prev=set()
        for q,a,w_ in self.delta_stepwise(w):
            filename_=os.path.join(dirpath,f'{i}')
            fin=len(w)-len(w_)
            if fin:
                processed=w[:fin-1]
            else:
                processed=" "
            a = a if a else " "
            if isinstance(q,set):
                g=self.save_img(filename_,q_c=q,a_c=set([a]),q_prev=q_prev,
                        symbols=symbols,states=states,
                        dpi=dpi,string=(processed,a,w_),format="png")
                q_prev=q
            else:
                g=self.save_img(filename_,q_c=set([q]),a_c=set([a]),q_prev=q_prev,
                        symbols=symbols,states=states,
                        dpi=dpi,string=(processed,a,w_),format="png")
                q_prev=set([q])
            im=Image.open(filename_+".png")
            images.append(im)
            i+=1
            filename_=os.path.join(dirpath,f'{i}')
            if isinstance(q,set):
                g=self.save_img(filename_,q_c=q,
                        symbols=symbols,dpi=dpi,string=(processed,a,w_),format="png")
            else:
                g=self.save_img(filename_,q_c=set([q]),
                        symbols=symbols,dpi=dpi,string=(processed,a,w_),format="png")
            im=Image.open(filename_+".png")
            if i==0 or len(w_)==0:
                images.append(im)
                images.append(im)
                images.append(im)
                images.append(im)
            images.append(im)
            i+=1

        images[0].save(filename,
                save_all=True, append_images=images[1:], optimize=False, duration=duration, loop=loop)
        if show:
            return HTML(f'<img src="{filename}">')
        else:
            return




# Código para reducciones

from maquinas.regular.dfa import DeterministicFiniteAutomaton as DFA
from maquinas.regular.ndfa import NonDeterministicFiniteAutomaton as NDFA
from maquinas.regular.ndfa_e import NonDeterministicFiniteAutomaton_epsilon as NDFA_e
from maquinas.regular.ndfa_e import epsilon
from maquinas.exceptions import *
from ordered_set import OrderedSet


def ndfa_e2ndfa(ndfa_e,rename=True,remove_sink=True):
    ndfa=NDFA()

    extra_A=set()
    for q_i in ndfa_e.Q:
        for a in ndfa_e.sigma:
            if a == epsilon:
                if len(ndfa_e.A.intersection(ndfa_e.expansion_epsilon(set([q_i]))))>0:
                    extra_A.add(q_i)
                continue
            qs_f=ndfa_e.delta_extended(set([q_i]),a)
            ndfa.add_transition(q_i,a,qs_f,force=True)

    ndfa.set_initial_state(ndfa_e.q_0)
    ndfa.A=ndfa_e.A.union(extra_A)
    if remove_sink:
        ndfa.remove_sink_states()
    if rename:
        ndfa.autorename()
    return ndfa

def codify_state(Q,s):
    s_=["0" for _ in Q]
    for q in s:
        s_[Q.index(q)]="1"
    return "".join(s_)


def ndfa2dfa(ndfa,rename=True,remove_sink=True):
    dfa=DFA()
    max_size=len(ndfa.Q)+1
    initial_state=set([ndfa.q_0])
    initial_state_=codify_state(ndfa.Q,initial_state)
    new_states={}
    new_states[initial_state_]=initial_state
    A=set()
    if len(initial_state.intersection(ndfa.A))>0:
        A.add(initial_state_)
    states={}
    lprev=-1
    while len(new_states)>0:
        states.update(new_states)
        lprev=len(states)
        new_new_states={}
        for state_,elements in new_states.items():
            for a in ndfa.sigma:
                new_state=set()
                for q in elements:
                    try:
                        new_state.update(ndfa.get_transition(q,a))
                    except DoesNotExistsTransition:
                        pass
                if len(new_state)==0:
                    continue
                new_state_=codify_state(ndfa.Q,new_state)
                dfa.add_transition(state_,a,set([new_state_]),force=True)
                if not new_state_ in states:
                    new_new_states[new_state_]=new_state
                if len(new_state.intersection(ndfa.A))>0:
                    A.add(new_state_)
        new_states=new_new_states
    dfa.set_initial_state(initial_state_)
    dfa.set_aceptors(A)

    if rename:
        dfa.autorename()
    if remove_sink:
        dfa.remove_sink_states()
    return dfa


def dfa2ndfa_e(dfa,rename=True,remove_sink=True):
    ndfa_e=NDFA_e()
    ndfa_e.Q=OrderedSet(dfa.Q)
    ndfa_e.sigma.update(dfa.sigma)
    for nq_i,t_ in dfa.ttable.items():
        for na,nq_f in t_.items():
            na=na+1
            try:
                ndfa_e.ttable[nq_i][na]=set(nq_f)
            except KeyError:
                ndfa_e.ttable[nq_i]={}
                ndfa_e.ttable[nq_i][na]=set(nq_f)

    ndfa_e.q_0=dfa.q_0
    ndfa_e.A=set(dfa.A)
    if rename:
        ndfa_e.autorename()
    if remove_sink:
        ndfa_e.remove_sink_states()
    return ndfa_e


def ndfa_e2dfa(ndfa_e,rename=True,remove_sink=True):
    return ndfa2dfa(ndfa_e2ndfa(ndfa_e,rename=rename,remove_sink=remove_sink),rename=rename,remove_sink=remove_sink)

def ndfa2ndfa_e(ndfa,rename=True,remove_sink=True):
    return dfa2ndfa_e(ndfa2dfa(ndfa,rename=rename,remove_sink=remove_sink),rename=rename,remove_sink=remove_sink)

def dfa2ndfa(dfa,rename=True,remove_sink=True):
    return ndfa_e2ndfa(dfa2ndfa_e(dfa,rename=rename,remove_sink=remove_sink),rename=rename,remove_sink=remove_sink)




cyjs_style = [
    {
        'selector': 'node',
        'style': {
            'border-color':'black',
            'background-color':'white',
            'border-width':2,
            'label': 'data(label)',
            'width': 50,
            'height': 50,
            'shape': 'circle',
            'color': 'black',
            'font-weight': 400,
            'text-halign': 'center',
            'text-valign': 'center',
            'font-size': 12
        }
    },
    {
        'selector': '.final',
        'style': {
            'border-width':5,
            'border-style':'double',
        }
    },
    {
        'selector': 'edge',
        'style': {
            'width': 2,
            'curve-style': 'bezier',
            'source-label':'data(label)',
            'line-color': 'black',
            'text-halign': 'center',
            'text-valign': 'top',
            'target-arrow-color': 'black',
            'target-arrow-shape': 'triangle',
            'source-text-offset': 60,
            'source-text-margin-y':-10,
            'target-endpoint': 'outside-to-node',
            'arrow-scale': 1,
        }
    }
]



# Código para reducciones

## TODO: https://arxiv.org/abs/1010.5318
## https://arxiv.org/pdf/1210.6624v1

from maquinas.regular.dfa import DeterministicFiniteAutomaton as DFA
from maquinas.regular.ndfa import NonDeterministicFiniteAutomaton as NDFA
from maquinas.regular.ndfa_e import NonDeterministicFiniteAutomaton_epsilon as NDFA_e
from maquinas.regular.ndfa_e import epsilon
from maquinas.exceptions import *
import random

def split(m,qs_f,a):
    res=set()
    for qf in qs_f:
        for q in m.Q:
            try:
                if qf in m.get_transition(q,a):
                    res.add(q)
            except DoesNotExistsTransition:
                pass
    return res

def forward(m,qs,a):
    res=set()
    for q in qs:
        try:
            qf=m.get_transition(q,a)
            qfs=list(qf)
            res.add(qfs[0])
        except DoesNotExistsTransition:
            pass
    return res



def minimization_hopcroft(dfa,rename=True,remove_sink=True):
    """ From http://www-igm.univ-mlv.fr/~berstel/Exposes/2009-06-08MinimisationLiege.pdf"""
    P_= [dfa.A, dfa.Q.difference(dfa.A)]
    W_ = []
    for a in dfa.sigma:
        W_.append((P_[0],a))
    while len(W_)>0:
        W,a=W_.pop(0)
        wsplit=split(dfa,W,a)
        P_new=[]
        for P in P_:
            if len(P.intersection(wsplit))==0 or P.intersection(wsplit) == P:
                P_new.append(P)
                continue

            P_1=wsplit.intersection(P)
            P_2=set(P)
            for p in P_1:
                P_2.remove(p)
            P_new.append(P_1)
            P_new.append(P_2)
            for b in dfa.sigma:
                if (P,b) in W_:
                    W_.remove((P,b))
                    W_.append((P_1,b))
                    W_.append((P_2,b))
                else:
                    if len(P_1)>len(P_2):
                        W_.append((P_2,b))
                    else:
                        W_.append((P_1,b))
        P_=P_new

    dfa_new=DFA(sigma=dfa.sigma)
    A=[]
    original2new={}
    for qs_i in P_:
        for q in qs_i:
            original2new[q]=qs_i

    for qs_i in P_:
        qs_il=[q for q in qs_i]
        qs_il.sort()
        qs_il=" ".join(qs_il)
        if dfa.q_0 in qs_i:
            initial=qs_il
        try:
            dfa_new.add_state(qs_il)
        except AlreadyExistsState:
            pass
        if len(dfa.A.intersection(qs_i))>0:
            A.append(qs_il)
        for a in dfa.sigma:
            qs_f=forward(dfa,qs_i,a)
            if qs_f:
                tmp=set()
                for q in qs_f:
                    tmp.update(original2new[q])
                qs_fl=[q for q in tmp]
                qs_fl.sort()
                qs_fl=" ".join(qs_fl)
                try:
                    dfa_new.add_state(qs_fl)
                except AlreadyExistsState:
                    pass
                dfa_new.add_transition(qs_il,a,[qs_fl])

    dfa_new.set_initial_state(initial)
    dfa_new.set_aceptors(A)

    if rename:
        dfa_new.autorename()

    if remove_sink:
        dfa_new.remove_sink_states()

    return dfa_new




# Base class for Finite Machines
from maquinas.exceptions import *
from ordered_set import OrderedSet
import re
import tempfile
import os
from collections import defaultdict

from PIL import Image
from IPython.core.display import display, HTML
from graphviz import Digraph

re_tape=re.compile(r'(\[B\]|epsilon|"[^"]+"|\w)')

class TuringMachine():
    """Turing machine"""

    def __init__(self, Q=[], sigma=[], gamma=[], B='𝖁',q_0=None,A=[], delta={}, force=False):
        """Common class for Turing Machine

        :param Q: Ordered set of states (default is empty).
        :param sigma: Ordered set of terminal symbols (default is empty).
        :param gamma: Ordered set of tape symbols (default is empty).
        :param B: Blank symbol (default 𝖁).
        :param q_0: Initial state (default None).
        :param A: Set of acceptor states (default is empty).
        :param delta: List of transitions with the form tupla of tupla of q_i and a, and the list of q_f, a symbol and direction (default is empty).
        :param force: If True and states or symbols do not exists create them (default is False).
        :type force: bool
        """
        self.sigma=OrderedSet()
        self.B=B
        self.i=0
        self.gamma=OrderedSet()
        self.gamma.add(self.B)
        self.gamma.update([self._filter(g) for g in gamma])
        self.gamma.update(sigma)
        self.sigma.update(sigma)
        self.Q=OrderedSet(Q)
        self.set_initial_state(q_0)
        self.set_aceptors(A,force=force)
        self.ttable={}
        for (q_i,a),qs in delta:
            self.add_transition(q_i,self._filter(a),[(q_f,self._filter(a),self._ndir(Dir)) for q_f,a,Dir in qs])
        self.curr=0
        self.tape=((),(self.B))

    def __getitem__(self,key):
        q,a=key
        return self.get_transition(q,a)

    def _nstate(self,q):
        return self.Q.index(q)

    def _nsymbol(self,a):
        return self.sigma.index(a)

    def _filter(self,t):
        if t == '[B]':
            return self.B
        elif t.lower() == 'blank':
            return self.B

        return t

    def _dir(self,d):
        if d == 1:
            return 'R'
        elif d == -1:
            return 'L'
        else:
            return 'N'

    def _ndir(self,d):
        if d == 'R':
            return 1
        elif d == 'L':
            return -1
        else:
            return 0

    def _ntsymbol(self,t):
        try:
            return self.gamma.index(self._filter(t))
        except KeyError as c:
            raise DoesNotExistsSymbol(c)

    def _state(self,nq):
        return self.Q.items[nq]

    def _symbol(self,na):
        return self.sigma.items[na]

    def _tsymbol(self,nz):
        return self.gamma.items[nz]

    def _status(self,status,states={},symbols={}):
        return "|".join(f"{states.get(s,s)}" for s,_,_,_ in status)

    def states(self):
        """ Gets states

        :returns: States of machine
        :rtype: list"""
        return list(self.Q)

    def symbols(self):
        """ Gets terminal symbols

        :returns: Terminal symbols of machine
        :rtype: list"""
        return list(self.sigma)

    def tsymbols(self):
        """ Gets tape symbols

        :returns: Tape symbols of machine
        :rtype: list"""
        return list(self.gamma)

    def tokens(self,r):
        """ Gets  tokens for tape

        :returns: Tape symbols for machine
        :rtype: list"""
        return re_tape.findall(r)

    def _transition(self,nq_i,nt,nq_f,nt_,Dir):
        return (self._state(nq_i),self._tsymbol(nt)),(self._state(nq_f),self._tsymbol(nt_),Dir)

    def __setitem__(self,key,value):
        q,t=key
        q_f,t_,Dir=value
        return self.add_transition(q,t,q_f,t_,Dir)

    def _get_transition(self,nq,nt):
        try:
            return self.ttable[nq][nt]
        except KeyError:
            return set()

    def delta(self,states):
        """ Applies delta function

        :param states: Internal state composed by (state, tape negative positions, tape positive positions, and (current position, symbol)
        :param a: Symbol

        :returns: Destination state"""
        states_=set()
        for nq,tn,tp,(c,a) in states:
            if c<0:
                try:
                    na=tn[-c]
                except IndexError:
                    tn=tn+tuple(self.B for _ in range(len(tn),1-c))
                    na=tn[-c]
            else:
                try:
                    na=tp[c]
                except IndexError:
                    tp=tp+tuple(self.B for _ in range(len(tp),c+1))
                    na=tp[c]
            qs=self._get_transition(nq,na)
            for nq_f,t_,Dir in qs:
                if c<0:
                    tn=list(tn)
                    tn[c]=t_
                elif c>=0:
                    tp=list(tp)
                    tp[c]=t_
                c=c+Dir
                if c<0 and len(tn)<=abs(c):
                    tn.append(self._ntsymbol(self.B))
                if c>=0 and len(tp)<=abs(c):
                    tp.append(self._ntsymbol(self.B))
                states_.add((nq_f,tuple(tn),tuple(tp),(len(tn)+c,self._tsymbol(na))))
        return states_

    def _tape(self,nt,pt):
        return [t for t in reversed(tuple(nt))]+[t for t in pt]

    def delta_stepwise(self,w,q=None,max_steps=0):
        """ Applies a step of delta extended function

        :param w: String
        :param q: Internal state where to start (default is initial state)
        :param max_steps: Maximun number of steps to consider

        :returns: Tuple with state of precessing at step, consisting of: state, left tape, right tape, (position, processed string)"""
        if q is None:
            states=set([(self._nstate(self.q_0),(),tuple(self._ntsymbol(a) for a in w),(0,""))])
            yield self._index2label(states)
        steps=0
        A=set(self._nstate(a) for a in self.A)
        while len(states)>0 and len(set([q for q,_,_,_ in states]).intersection(self.A))==0:
            states=self.delta(states)
            steps+=1
            yield self._index2label(states)
            if set([s for s,_,_,_ in states]).intersection(A):
                break
            if max_steps and steps>=max_steps:
                break

    def create_initial_istate(self,w):
        """ Creates an initial internal state 

        :param w: string for the tape

        :return: Returns internal initial state"""
        return set([(self._nstate(self.q_0),(),tuple(self._ntsymbol(a) for a in w),(0,""))])

    def delta_extended(self,states,w,max_steps=None):
        """ Applies delta extended function

        :param states: Internal states
        :param w: String
        :param max_steps: Maximum number of states

        :returns: Returns internal state after processing the full string"""

        if states is None:
            states=self.create_initial_istate(w)
        steps=0
        while len(states)>0 and len(set([self._state(q) for q,_,_,_ in states]).intersection(self.A))==0:
            states=self.delta(states)
            steps+=1
            res=[(q,tn,tp,c) for q,tn,tp,c in states]
            if max_steps and steps>=max_steps:
                return []
        return self._index2label(res)

    def _index2label(self,states_):
        return [(self._state(q),
            tuple(self._tsymbol(s) for s in l),
            tuple(self._tsymbol(s) for s in r),
            c
            )
            for q,l,r,c in states_]

    def _label2index(self,states):
        return [(self._nstate(q),
            tuple(self._ntsymbol(s) for s in l),
            tuple(self._ntsymbol(s) for s in r),
            c
            )
            for q,l,r,c in states]


    def items(self):
        """ Iterator over the transitions

        :returns: Yeilds a tuple transition"""
        for nq_i,val in self.ttable.items():
            for nt,nq_fs in val.items():
                for (nq_f,nt_,Dir) in nq_fs:
                    yield self._transition(nq_i,nt,nq_f,nt_,Dir)

    def step(self,states):
        return self._index2label(self.delta(states))

    def get_transition(self,q,a):
        """ Gets the destintion state or states for state, terminal symbol and stack symbol

        :param q: Source state
        :param a: Terminal symbol
        :returns: Destination state or states"""

        qs=self._get_transition(self._nstate(q),self._nsymbol(a))
        return [ (self._state(s),self.tsymbol(a),d) for s,a,d in qs]

    def add_transition(self,q_i,t,qs,force=False):
        """ Adds a transition

        :param q_i: Source state
        :param t: Tape symbol
        :param q_s: Destination state (q_f,t_2,dir)
        :param force: Force creation of elements
        :returns: None"""
        try:
            nq_i=self.add_state(q_i)
        except AlreadyExistsState:
            nq_i=self._nstate(q_i)
        try:
            nt=self.add_tsymbol(t)
        except AlreadyExistsSymbol:
            nt=self._ntsymbol(t)
        qs=[(self._nstate(q),self._ntsymbol(t),Dir) for q,t,Dir in qs]

        if nq_i in self.ttable and \
            nt in self.ttable[nq_i]:
                raise AlreadyExistsTMTransition(q_i,t,self)

        if not nq_i in self.ttable:
            self.ttable[nq_i]={}
        if not nt in self.ttable[nq_i]:
            self.ttable[nq_i][nt]=set()
        self.ttable[nq_i][nt].update(qs)

    def add_state(self,q,initial=False):
        """ Adds a state

        :param q: State or states
        :param initial: Set state as a initial
        :returns: Indixes of state or states"""
        if initial:
            self.q_0=q
        if q in self.Q:
            raise AlreadyExistsState(q)
        if isinstance(q,(set,list)):
            return set(self.Q.add(q_) for q_ in q)
        else:
            return self.Q.add(q)

    def add_next_state(self,initial=False):
        """ Adds a state with a infered name based on the number of states q_max. If the name state is already defined it looks the following integer available.

        :param q: State or states
        :param initial: Set state as a initial
        :returns: Next state generated and integer"""
        max_ix=len(self.Q)
        while f"q_{max_ix}" in self.Q:
            max_ix+=1
        q=f"q_{max_ix}"
        self.Q.add(q)
        if initial:
            self.q_0=q
        return q,max_ix

    def add_symbol(self,a):
        """ Adds a symbol

        :param a: Symbol
        :returns: Indixes of symbol"""
        if a in self.gamma:
            raise AlreadyExistsSymbol(a)
        return self.sigma.add(a)
        return self.gamma.add(a)

    def add_tsymbol(self,t):
        """ Adds a tape symbol

        :param t: Tape xymbol
        :returns: Indixes of symbol"""
        if t in self.gamma:
            raise AlreadyExistsSymbol(t)
        return self.gamma.add(t)

    def set_initial_state(self,q,force=False):
        """ Sets an initial state

        :param q: State
        :param force: If not defined it creates it (default is False)
        :returns: None"""
        if q is None:
            self.q_0=None
            return None
        if not q in self.Q:
            if force:
                self.add_state(q)
            else:
                raise DoesNotExistsState(q)
        self.q_0=q

    def get_initial_state(self):
        """ Gets an initial state

        :returns: State"""
        return self.q_0

    def set_aceptors(self,A,force=False):
        """ Sets aceptors states

        :param A: States
        :param force: If not defined it creates it (default is False)
        :returns: None"""
        if force:
            self.add_state(A)
        self.A=set(A)

    def stepStatus(self,status):
        """ Gives a step and calculates new status for Simulation

        :param Status: Status
        :returns: None"""
        if status.state is None:
            states=self._index2label(self.create_initial_istate(status.string))
        else:
            states=status.state
        states=self.step(self._label2index(states))
        status.position+=1
        status.step+=1
        status.state=states

    def accepts(self,w,max_steps=0):
        """ Checks if string is accepted

        :param w: String
        :returns: None"""
        try:
            return self.acceptor(self.delta_extended(None,w,max_steps=max_steps))
        except:
            return False

    def acceptor(self,states):
        """ Checks if state is an acceptor state

        :param states: State or states
        :type: Set

        :returns: None"""
        final=set([q for q,_,_,_ in states])
        if bool(final.intersection(self.A)):
            return True
        return False

    def save_img(self,filename,q_c=set(),a_c=set(),q_prev=set(),symbols={},states={},format='svg',dpi="60.0",string=None,status=None):
        """ Saves machine as an image

            :param filename: Filename of image
            :param q_c: Set of current states to be highlited (default is empty)
            :param a_c: Set of current symbols to be highlited (default is empty)
            :param q_prev: Set of previos states to be highlited (default is empty)
            :param symbols: Replacements of the symbols to show (default is empty)
            :param states: Replacements of the states to show (default is empty)
            :param format: Format of image (default is svg)
            :param dpi: Resolution of image (default is "90.0")
            :param string: Label of string being analysed (default is None)
            :param stack: Status of the stack (default is None)
            :param status: Status of the TM (default is None)

            :returns: None"""
        dot=self.graph(q_c=q_c,a_c=a_c,q_prev=q_prev,symbols=symbols,states=states,format=format,dpi=dpi,string=string,status=status)
        dot.render(filename,format="png",cleanup=True)

    def states2string(self,states):
        """ Renders srting with the state of the TM

        :returns: String tiwh the states of the TM"""
        res=[]
        for q,nt,pt,(c,a) in states:
            tape=self._tape(nt,pt)
            res.append("{}, … {} _{}_ {} …".format(q,\
                "".join([ t for t in tape[:c]]),\
                tape[c],\
                "".join([ t for t in tape[c+1:]])))
        return "|".join(res)

    def summary(self):
        """ Producrs summary of the PDA
        :returns: List with summary"""
        info= [
         "States  : "+", ".join(self.states()),
         "Sigma   : "+", ".join(self.symbols()),
         "Gamma   : "+", ".join(self.tsymbols()),
         "Initial : "+self.q_0,
         "Aceptors: "+", ".join(self.A),
         "Transitions:\n"+"\n".join(f" {q_i},{t}/{t_} → {q_f},{Dir}" for (q_i,t),(q_f,t_,Dir) in self.items())]
        return "\n".join(info) 

    def print_summary(self):
        """ Prints a summary of the PDA
        """
        print(self.summary())


    def graph(self,q_c=set(),a_c=set(),q_prev=set(),symbols={},states={},format="svg",dpi="60.0",string=None,status=None,one_arc=True,finished=False):
        """ Graphs TM

        :param q_c: Set of current states to be highlited (default is empty)
        :param a_c: Set of current symbols to be highlited (default is empty)
        :param q_prev: Set of previos states to be highlited (default is empty)
        :param symbols: Replacements of the symbols to show (default is empty)
        :param states: Replacements of the states to show (default is empty)
        :param format: Format of image (default is svg)
    :param dpi: Resolution of image (default is "60.0")
        :param string: Label of string being analysed (default is None)
        :param status: Status of the TM (default is None)
        :param one_arc: Graph one arc in case of multiple transitions (default is True)
        :param finished: If has pass through final state (default is False)

        :returns: Returns Digraph object from graphviz"""
        if len(q_c)==0:
            states_=[(self._nstate(self.q_0),(),(),(0,""))]
        else:
            states_=q_c

        f=Digraph(comment="TM",format=format)
        f.attr(rankdir='LR',dpi=dpi)
        for i,(q_c_,nt,pt,(c,a)) in enumerate(states_):
            if len(q_c)>0:
                q_c_=set([q_c_])
            else:
                q_c_=[]
            with f.subgraph(name=f'cluster_{i}') as f_:
                self._graph(f_,
                    i=i,
                    q_c=q_c_,
                    a_c=set([a]),
                    q_prev=q_prev,
                    symbols=symbols,
                    states=states,
                    tape=self._tape(nt,pt),
                    pos=c+len(nt),
                    dpi=dpi,
                    format=format,
                    status=status,
                    string=string,
                    one_arc=one_arc)
        return f

    def _graph(self,f,i=0,q_c=set(),a_c=set(),q_prev=set(),states={},symbols={},format="svg",dpi="60.0",string=None,tape=None,status=None,one_arc=True,pos=None,finished=False):
        label_tape=None
        if len(self.A.intersection(q_c))>0:
            color_state="limegreen"
        else:
            if status==None:
                color_state="lightblue2"
            else:
                color_state="orangered"

        f.attr(style='invis',labelloc="b")

        if tape:
            cells=[]
            for i,c in enumerate(tape):
                if i==pos:
                    cells.append(f'<TD BGCOLOR="{color_state}">{symbols.get(c,c)}</TD>')
                else:
                    cells.append(f'<TD>{symbols.get(c,c)}</TD>')
            label_tape=f"< <TABLE BORDER='0' CELLBORDER='1' SIDES='TBRL'><TR>{' '.join(cells)}</TR></TABLE> >"

        if label_tape:
            f.attr(label=label_tape)

        for q,_ in self.Q.map.items():
            if q in self.A:
                shape="doublecircle"
            else:
                shape="circle"
            if q in q_c:
                f.node(name=f'{q}_{i}',label=states.get(q,q),shape=shape,color=color_state,style="filled")
            else:
                f.node(name=f'{q}_{i}',label=states.get(q,q),shape=shape)

        edges=defaultdict(list)
        for e,info in enumerate(self.items()):
            (q_i,a),(q_f,a_,Dir) = info
            if (q_f in q_c and q_i in q_prev) and (a in a_c):
                edges[(f'{q_i}_{i}',f'{q_f}_{i}')].append((f'{symbols.get(a,a)}/{symbols.get(a_,a_)},{self._dir(Dir)}',True))
            else:
                edges[(f'{q_i}_{i}',f'{q_f}_{i}')].append((f'{symbols.get(a,a)}/{symbols.get(a_,a_)},{self._dir(Dir)}',False))

        for (q_i,q_f),labels in edges.items():
            if one_arc:
                tags=[]
                colored_=False
                for label,colored in labels:
                    if colored:
                        colored_=True
                        tags.append(f'<FONT color="{color_state}">{label}</FONT>')
                    else:
                        tags.append(f'{label}')
                tags=f'< {"<BR/>".join(tags)} >'
                if colored_:
                    f.edge(q_i,q_f,label=tags,labelloc='b',color=color_state)
                else:
                    f.edge(q_i,q_f,label=tags,labelloc='b')
            elif not one_arc:
                for label,colored in labels:
                    if colored:
                        f.edge(q_i,q_f,label=label,labelloc='b',color=color_state,fontcolor=color_state)
                    else:
                        f.edge(q_i,q_f,label=label,labelloc='b')
        return f


    def table(self,symbols={},states={},q_order=None,s_order=None,color_final="#32a852",empty_symbol="∅"):
        """ Creates an HTML object for the table of the PDA

        :param symbols: Replacements of the symbols to show (default is empty)
        :param states: Replacements of the states to show (default is empty)
        :param  q_order: Order to use for states
        :param  s_order: Order to use for symbols
        :param color_final: RGB string for color of final state (default is "#32a852")

        :returns: Display object for IPython"""
        if not s_order:
            s_order=list(self.gamma)
            s_order.sort()
        if not q_order:
            q_order=list(self.Q)
            q_order.sort()
        symbs_h="</strong></td><td><strong>".join([symbols.get(q,q) for q in s_order])
        table=f"<table><tr><td></td><td><strong>{symbs_h}</strong></td></tr>"
        for q_i in q_order:
            vals=[]
            initial="⟶" if q_i == self.q_0 else ""
            final=f'bgcolor="{color_final}"' if q_i in self.A else ""
            vals.append(f"<strong>{initial}{states.get(q_i,q_i)}</strong>")
            for a in s_order:
                try:
                    labels=[]
                    for q_f,r,Dir in self.ttable[self._nstate(q_i)][self._ntsymbol(a)]:
                        labels.append(f'/{symbols.get(self._tsymbol(r),self._tsymbol(r))}→{self._state(q_f)},{self._dir(Dir)}')
                    vals.append("<br/>".join(labels))
                except KeyError:
                    vals.append(empty_symbol)
            row="</td><td>".join(vals)
            table+=f"<tr><td {final}>{row}</td></tr>"
        table+="</table>"
        return display(HTML(table))

    def save_gif(self,w,filename="tm.gif",symbols={},states={},dpi="90",show=True,loop=0,duration=500,max_steps=1000):
        """ Saves an animation of  machine

        :param w: String to analysed during animation
        :param filename: Name of gif (default is tm.gif")
        :param symbols: Replacements of the symbols to show (default is empty)
        :param states: Replacements of the states to show (default is empty)
        :param dpi: Resolution of image (default is "90.0")
        :param show: In interactive mode return gif
        :param loop: Number of loops in annimation, cero is forever (default is 0)
        :param duration: Duration in msegs among steps (default is 500)
        :param max_steps: Maximum number of steps to consider (default is 1000)
        :returns: None or HTML for Ipython"""
        dirpath = tempfile.mkdtemp()
        i=0
        images=[]
        q_prev=set()
        max_images_height=1
        status=None
        for ii,q in enumerate(self.delta_stepwise(w)):
            if len(q)==0:
                status=self.accept(q_prev)
                if status:
                    break
                q=q_prev
                q_prev=set()
            if ii>=max_steps:
                break
            filename_=os.path.join(dirpath,f'{i}')
            g=self.save_img(filename_,q_c=q,q_prev=set([q_c for q_c,_,_,_ in q]),
                    symbols=symbols,states=states,status=status,
                    dpi=dpi,format="png")
            q_prev=q
            im=Image.open(filename_+".png")
            width, height = im.size
            max_images_height=max(max_images_height,height)
            images.append(im)
            i+=1
            filename_=os.path.join(dirpath,f'{i}')
            g=self.save_img(filename_,q_c=q,status=status,
                        symbols=symbols,states=states,dpi=dpi,format="png")
            im=Image.open(filename_+".png")
            images.append(im)
            i+=1
        images.append(im)
        images.append(im)
        images.append(im)
        images.append(im)

        for i,im in enumerate(images):
            im2 = Image.new('RGB', (width, max_images_height), (255, 255, 255))
            width, height = im.size
            im2.paste(im)
            images[i]=im2

        images[0].save(filename,
                save_all=True, append_images=images[1:], optimize=False, duration=500, loop=loop)
        if show:
            return HTML(f'<img src="{filename}">')
        else:
            return




# Código de simulación
import ipywidgets as widgets
from IPython.core.display import display, HTML, Markdown, clear_output
from maquinas.status import Status

class SimulationGrammar():
    def __init__(self,grammar,string="",speed=1,symbols={},states={}):
        self.speed=speed
        self.symbols=symbols
        self.states=states
        self.input_string = widgets.Text(
            value=string,
            description='Cadena', )
        self.button_load_string = widgets.Button(description='⏎')
        self.string=widgets.HBox( [self.input_string,
            self.button_load_string])
        self.button_forward = widgets.Button(description='▶️')
        self.button_reverse = widgets.Button(description='◀️')
        self.controls=widgets.HBox( [self.button_reverse,
            self.button_forward])
        self.output = widgets.Output()
        self.button_forward.on_click(self._next)
        self.button_load_string.on_click(self._load_string)

    def reset(self,strin=""):
        self.status=Status(self.status.machine)

    def display(self):
        with self.output:
            clear_output()
            display(HTML(self.status.tape2html(symbols=self.symbols,states=self.states)))
            display(self.status.machine.graph(self.status.state,symbols=self.symbols,states=self.states))

    def _load_string(self,b):
        self.reset()
        self.status.tape=self.input_string.value
        self.display()

    def _next(self,b):
        self.status.next()
        self.display()

    def run(self,text=None):
        self.status.tape=self.input_string.value
        self.display()
        return display(
                widgets.VBox(
                    [self.string,
                        self.controls,
                        self.output]))





# Eceptions for library

class AlreadyExistsState(Exception):
    """Already exists state"""
    def __init__(self, q):
        self.q = q
        msg= f"State '{q}' already exists in machine"
        super().__init__(msg)

class AlreadyExistsSymbol(Exception):
    """Already Exists Symbol"""
    def __init__(self, a):
        self.a = a
        msg= f"Symbol '{a}' already exists in machine"
        super().__init__(msg)

class AlreadyExistsTransition(Exception):
    """Already exists transition"""
    def __init__(self, q_i, a,m):
        self.a = a
        self.q_i = q_i
        self.m = m
        msg= f"Transition ({q_i},{a}) already exists with destination {m[q_i,a]}"
        super().__init__(msg)

class AlreadyExistsTMTransition(Exception):
    """Already exists transition Turing Machine"""
    def __init__(self, q_i, a,m):
        self.a = a
        self.q_i = q_i
        self.m = m
        msg= f"Transition ({q_i},{a}) already exists with destination {m.get_transition(q_i,a)}"
        super().__init__(msg)

class AlreadyExistsTSPDATransition(Exception):
    """Already exists transition two stack"""
    def __init__(self, q_i, a, z1, z2, m):
        self.a = a
        self.q_i = q_i
        self.m = m
        self.z1 = z1
        self.z2 = z2
        msg= f"Transition ({q_i},{a},{z1},{z2}) already exists with destination {m.get_transition(q_i,a,z1,z2)}"
        super().__init__(msg)

class AlreadyExistsPDATransition(Exception):
    """Already exists transition"""
    def __init__(self, q_i, a, z, m):
        self.a = a
        self.q_i = q_i
        self.m = m
        self.z = z
        msg= f"Transition ({q_i},{a},{z}) already exists with destination {m.get_transition(q_i,a,z)}"
        super().__init__(msg)

class DoesNotExistsTransition(Exception):
    """Does not exists transition"""
    def __init__(self, q_i, a):
        self.a = a
        self.q_i = q_i
        msg= f"Transition ({q_i},{a}) is not defined in machine"
        super().__init__(msg)

class DoesNotExistsState(Exception):
    """Does not exists state"""
    def __init__(self, q):
        self.q = q
        msg= f"State {q} is not defined in machine"
        super().__init__(msg)

class DoesNotExistsSymbol(Exception):
    """Does not exists symbol defined"""
    def __init__(self, a):
        self.a = a
        msg= f"Symbol '{a}' is not defined in machine"
        super().__init__(msg)

class NoIntitialStateDefined(Exception):
    """There is not initial state defined"""
    def __init__(self):
        msg= f"There is not initial state defined"
        super().__init__(msg)

class NoStringWithDefinition(Exception):
    """No string with definition of RE or grammar"""
    def __init__(self, s, t):
        self.s = s
        self.t = t
        msg= f"The {t} was provided with an empty definition '{s}'"
        super().__init__(msg)

class NoSymbolInMachine(Exception):
    """No symbol in the definition in the machine"""
    def __init__(self, s, ts):
        self.s = s
        self.ts = ts
        msg= f"The {s} symbol is not part of the terminals: {','.join(ts)}"
        super().__init__(msg)






# Base class for Finite Machines
from maquinas.exceptions import *
from collections import defaultdict
from ordered_set import OrderedSet
import re
import tempfile
import os

from PIL import Image
from IPython.core.display import display, HTML
from graphviz import Digraph

re_queque=re.compile(r'(Z\d+|epsilon|_[^_]+_|\w)')

class TwoStackPushDownAutomaton():
    """Common class for Two stack Push Down Automaton

    :param Q: Ordered set of states (default is empty).
    :param sigma: Ordered set of terminal symbols (default is empty).
    :param gamma: Ordered set of stack symbols (default is empty).
    :param q_0: Initial state (default None).
    :param Z_0: Initial stack symbol (default None).
    :param A: Set of acceptor states (default is empty).
    :param delta: List of transitions with the form tupla of tupla of q_i and a, and q_f (default is empty).
    :param force: If True and states or symbols do not exists create them (default is False).
    :type force: bool
    """
    def __init__(self, Q=[], sigma=[], gamma=[], q_0=None,Z_0=None,A=[], delta={}, force=False):
        self.sigma=OrderedSet()
        self.sigma.add('ε')
        self.gamma=OrderedSet()
        self.gamma.add('ε')
        Z_0=self.set_initial_qsymbol(Z_0)
        self.gamma.add(Z_0)
        self.sigma.update(sigma)
        self.gamma.update([self._filter(g) for g in gamma])
        self.Q=OrderedSet(Q)
        self.set_initial_state(q_0)
        self.set_aceptors(A,force=force)
        self.ttable={}
        for (q_i,a,z1,z2),qs in delta:
            # Replace Z_0 or Z0 by Z₀ or epsilon for ε
            self.add_transition(q_i,self._filter(a),self._filter(z1),self._filter(z2),self._filter_stack(qs))

    def _filter_stack(self,qs):
        return [(q,(tuple(self._filter(t) for t in self.tokens(r1)),tuple(self._filter(t) for t in self.tokens(r2)))) for (q,r1,r2) in qs]

    def __getitem__(self,key):
        q,a,z1,z2=key
        return self.get_transition(q,a,z1,z2)

    def _nstate(self,q):
        return self.Q.index(q)

    def _nsymbol(self,a):
        return self.sigma.index(self._filter(a))

    def _filter(self,z):
        if z in ['Z0','Z_0']:
            z='Z₀'
        if z in ['epsilon']:
            z='ε'
        return z

    def _nqsymbol(self,z):
        return self.gamma.index(self._filter(z))

    def _state(self,nq):
        return self.Q.items[nq]

    def _symbol(self,na):
        return self.sigma.items[na]

    def _qsymbol(self,nz):
        return self.gamma.items[nz]

    def _status(self,status,states={},symbols={}):
        return "|".join(f"{states.get(q,q)},\
            {''.join(symbols.get(z,z) for z in r1)}]①  {''.join(symbols.get(z,z) for z in r1)}]②" for q,(r1,r2) in status)

    def states(self):
        """ Gets states

        :returns: States of machine
        :rtype: list"""
        return list(self.Q)

    def symbols(self):
        """ Gets terminal symbols

        :returns: Terminal symbols of machine
        :rtype: list"""
        return list(self.sigma)

    def qsymbols(self):
        """ Gets stack symbols

        :returns: Stack symbols of machine
        :rtype: list"""
        return list(self.gamma)

    def tokens(self,r):
        """ Gets  tokens for stack

        :returns: Stack symbols for machine
        :rtype: list"""
        return re_queque.findall(r)

    def _transition(self,nq_i,na,nz1,nz2,nq_f,nq):
        return (self._state(nq_i),self._symbol(na),self._qsymbol(nz1),self._qsymbol(nz2)),(self._state(nq_f),tuple(tuple(self._qsymbol(n) for n in nr) for nr in nq))

    def __setitem__(self,key,value):
        q,a,z1,z2=key
        q_f,qs_1,q_s2=value
        return self.add_transition(q,a,z1,z2,q_f,(qs_1,q_s2))

    def _get_transition(self,nq,na,nz1,nz2):
        try:
            return self.ttable[nq][na][nz1][nz2]
        except KeyError:
            return set()

    def expansion_epsilon(self,qs):
        """ Applies expansion by epsilon in a set of states

        :param qs: Set of states

        :returns: Set of reachable states from qs"""
        qs__=defaultdict(set)
        qs__.update(qs)
        change=True
        expanded=False
        while change:
            qs_=defaultdict(set)
            qs_.update(qs__)
            qs_.update(self.delta(qs__,'ε',index=True))
            if len(set(qs__.keys()) ^ set(qs_.keys()))==0:
                change=False
            else:
                expanded=True
            qs__=qs_
        return qs__,expanded

    def delta_extended(self,states,w,index=False):
        """ Applies delta extended function

        :param q: Internal state
        :param w: String
        :param index: If returns indexes instead of labesl (default False)

        :returns: Returns internal state after processing the full string"""
        if states is None:
            states=self.create_initial_istate()
        if len(w)==0:
            res,_=self.expansion_epsilon(states)
            if index:
                return res
            else:
                return self._index2label(res)
        else:
            *w_,a=w
            q_u=defaultdict(set)
            for k,v in self.delta_extended(states,w_,index=True).items():
                r_={k:v}
                q_u.update(self.delta(r_,a,index=True))
            res,_ = self.expansion_epsilon(q_u)
            if index:
                return res
            else:
                return self._index2label(res)

    def _index2label(self,states_):
        return dict([((self._state(nq_f),(tuple(self._qsymbol(z) for z in r[0]),tuple(self._qsymbol(z) for z in r[1]))),set(tuple(self._qsymbol(n) for n in z) for z in nz)) for (nq_f,r), nz in states_.items()])

    def _label2index(self,states_):
        return dict([((self._nstate(nq_f),(tuple(self._nqsymbol(z) for z in r[0]),tuple(self._nqsymbol(z) for z in r[1]))),set(tuple(self._nqsymbol(n) for n in z) for z in nz)) for (nq_f,r), nz in states_.items()])

    def create_istates(self,states):
        """ Creates a internal state for de PDA 

        :param states: List of tuples (state, stack)

        :returns: Usable internal state for the PDA"""

        return dict([((self._nstate(q),tuple(tuple(self._nqsymbol(z) for z in q) for q in qs)),set()) for q,qs in states])

    def create_initial_istate(self):
        """ Creates initial internal state for de PDA 
        :returns: Usable internal state for the PDA"""
        return dict([((self._nstate(self.q_0),((self._nqsymbol(self.Z_0),),(self._nqsymbol(self.Z_0),))),set())])

    def delta(self,states,a,index=False):
        """ Applies delta function

        :param states: Internal state composed by (state, stack)
        :param a: Symbol

        :returns: Destination state"""
        states_=defaultdict(OrderedSet)
        na=self._nsymbol(self._filter(a))
        for nq,(ns1,ns2) in states:
            nz1=ns1[0]
            nz2=ns2[0]
            stack1=ns1[1:]
            stack2=ns2[1:]
            qs=self._get_transition(nq,na,nz1,nz2)
            for nq_f,(r1,r2) in qs:
                r1=tuple(r1)
                r2=tuple(r2)
                if len(r1)==1 and r1[0]==0:
                    s1_=stack1
                else:
                    s1_=r1+stack1
                if len(r2)==1 and r2[0]==0:
                    s2_=stack2
                else:
                    s2_=r2+stack2
                states_[(nq_f,(s1_,s2_))].add((nz1,nz2))
        if index:
            return states_
        else:
            return self._index2label(states_)

    def delta_stepwise(self,w,istates=None,mark_finished=False,index=False):
        """ Applies a step of delta extended function

        :param w: String
        :param istates: Internal state where to start (default is initial state)
        :param mark_finished: Mark if finished (default is False)
        :param index: If returns indexes instead of labesl (default False)

        :returns: Tuple with state of precessing at step, consisting of: internal state, procesed symbol and processed string"""
        if istates is None:
            istates=self.create_initial_istate()
        if mark_finished:
            yield self._index2label(istates),"",w,len(w)==0
        else:
            yield self._index2label(istates),"",w
        ix=0
        for ix,a in enumerate(w):
            istates=self.delta(istates,a,index=True)
            istates,expanded=self.expansion_epsilon(istates)
            if mark_finished:
                yield self._index2label(istates),a,w[ix+1:],(ix+1)==len(w)
            else:
                yield self._index2label(istates),a,w[ix+1:]

    def items(self):
        """ Iterator over the transitions

        :returns: Yeilds a tuple transition"""
        for nq_i,t_ in self.ttable.items():
            for na,t__ in t_.items():
                for nz1,t___ in t__.items():
                    for nz2,nq_fs in t___.items():
                        for (nq_f,nr) in nq_fs:
                            yield self._transition(nq_i,na,nz1,nz2,nq_f,nr)

    def step(self,states,a):
        res,_=self.expansion_epsilon(self.delta(states,a,index=True))
        return self._index2label(res)

    def get_transition(self,q,a,z1,z2):
        """ Gets the destintion state or states for state, terminal symbol and stack symbol

        :param q: Source state
        :param a: Terminal symbol
        :param z: Stack symbol
        :returns: Destination state or states"""
        qs=self._get_transition(self._nstate(q),self._nsymbol(a),self._nqsymbol(z1),self._nqsymbol(z2))
        return [(self._state(q),tuple(tuple(self._qsymbol(s) for s in r) for r in qr )) for q,qr in qs ]

    def add_transition(self,q_i,a,z1,z2,qs,force=False):
        """ Adds a transition

        :param q_i: Source state
        :param a: Terminal ymbol
        :param z1: Stack 1 symbol
        :param z2: Stack 2 symbol
        :param q_s: Destination state (q_f,q_1,q_2)
        :param forece: Force creation of elements
        :returns: None"""
        try:
            nq_i=self.add_state(q_i)
        except AlreadyExistsState:
            nq_i=self._nstate(q_i)
        try:
            na=self.add_symbol(self._filter(a))
        except AlreadyExistsSymbol:
            na=self._nsymbol(self._filter(a))
        try:
            nz1=self.add_qsymbol(self._filter(z1))
        except AlreadyExistsSymbol:
            nz1=self._nqsymbol(z1)
        try:
            nz2=self.add_qsymbol(self._filter(z2))
        except AlreadyExistsSymbol:
            nz2=self._nqsymbol(z2)

        if force:
            for q_f,rs in qs:
                try:
                    self.add_state(q_f)
                except AlreadyExistsState:
                    pass
                for r in rs:
                    for s in r:
                        try:
                            self.add_qsymbol(s)
                        except AlreadyExistsSymbol:
                            pass
        nqs=[(self._nstate(q),tuple(tuple(self._nqsymbol(s) for s in r) for r in qr )) for q,qr in qs ]
        if nq_i in self.ttable and \
           na in self.ttable[nq_i] and \
           nz1 in self.ttable[nq_i][na] and \
           nz2 in self.ttable[nq_i][na][nz1]:
               raise AlreadyExistsTSPDATransition(q_i,a,z1,z2,self)

        if not nq_i in self.ttable:
            self.ttable[nq_i]={}
        if not na in self.ttable[nq_i]:
            self.ttable[nq_i][na]={}
        if not nz1 in self.ttable[nq_i][na]:
            self.ttable[nq_i][na][nz1]={}
        if not nz2 in self.ttable[nq_i][na][nz1]:
            self.ttable[nq_i][na][nz1][nz2]=set()
        self.ttable[nq_i][na][nz1][nz2].update(nqs)

    def add_state(self,q,initial=False):
        """ Adds a state

        :param q: State or states
        :param initial: Set state as a initial
        :returns: Indixes of state or states"""
        if initial:
            self.q_0=q
        if q in self.Q:
            raise AlreadyExistsState(q)
        if isinstance(q,(set,list)):
            return set(self.Q.add(q_) for q_ in q)
        else:
            return self.Q.add(q)

    def add_next_state(self,initial=False):
        """ Adds a state with a infered name based on the number of states q_max. If the name state is already defined it looks the following integer available.

        :param q: State or states
        :param initial: Set state as a initial
        :returns: Next state generated and integer"""
        max_ix=len(self.Q)
        while f"q_{max_ix}" in self.Q:
            max_ix+=1
        q=f"q_{max_ix}"
        self.Q.add(q)
        if initial:
            self.q_0=q
        return q,max_ix

    def add_symbol(self,a):
        """ Adds a symbol

        :param a: Symbol
        :returns: Indixes of symbol"""
        if a in self.gamma:
            raise AlreadyExistsSymbol(self._filter(a))
        self.gamma.add(a)
        return self.sigma.add(a)

    def add_qsymbol(self,z):
        """ Adds a stack symbol

        :param a: Symbol
        :returns: Indixes of symbol"""
        z= self._filter(z)
        if z in self.gamma:
            raise AlreadyExistsSymbol(z)
        return self.gamma.add(z)

    def set_initial_state(self,q,force=False):
        """ Sets an initial state

        :param q: State
        :param force: If not defined it creates it (default is False)
        :returns: None"""
        if q is None:
            self.q_0=None
            return None
        if not q in self.Q:
            if force:
                self.add_state(q)
            else:
                raise DoesNotExistsState(q)
        self.q_0=q

    def set_initial_qsymbol(self,z=None,force=False):
        """ Sets an initial symbol for the  stack

        :param q: State
        :param force: If not defined it creates it (default is False)
        :returns: None"""
        if z is None:
            self.Z_0='Z₀'
        else:
            if force and not z in self.gamma:
                self.gamma.add(z)
            self.Z_0=z
        return self.Z_0

    def get_initial_state(self):
        """ Gets an initial state

        :returns: State"""
        return self.q_0

    def set_aceptors(self,A,force=False):
        """ Sets aceptors states

        :param A: States
        :param force: If not defined it creates it (default is False)
        :returns: None"""
        if force:
            self.add_state(A)
        self.A=set(A)

    def accepts(self,w):
        """ Checks if string is accepted

        :param w: String
        :returns: None"""
        return self.acceptor(set([q for q,_ in self.delta_extended(None,w).keys()]))

    def acceptor(self,q):
        """ Checks if state is an acceptor state

        :param q: State or states
        :type: Set

        :returns: None"""
        if isinstance(q,dict):
            q=set([q for q,_ in q.keys()])
            if bool(q.intersection(self.A)):
                return True
        if isinstance(q,list):
            q=set([q for q,_ in q])
            if bool(q.intersection(self.A)):
                return True
        elif isinstance(q,set):
            if bool(q.intersection(self.A)):
                return True
        else:
            if q in self.A:
                return True
        return False


    def stepStatus(self,status):
        """ Gives a step and calculates new status for Simulation

        :param Status: Status
        :returns: None"""
        if status.state is None:
            states=self._index2label(self.create_initial_istate())
        else:
            states=status.state

        a=status.get_symbol_tape()
        states=self.step(self._label2index(states),a)
        status.position+=1
        status.step+=1
        status.state=states

    def states2string(self,states):
        """ Renders srting with the state of the TSPDA

        :returns: String tiwh the states of teh TSPDA"""
        return " | ".join(["{}, {}]① {}]②".format(s, " ".join(r1), " ".join(r2)) for s,(r1,r2) in states ])

    def summary(self):
        """ Producrs summary of the PDA
        :returns: List with summary"""
        info= [
         "States  : "+", ".join(self.states()),
         "Sigma   : "+", ".join(self.symbols()),
         "Gamma   : "+", ".join(self.qsymbols()),
         "Initial : "+self.q_0,
         "Aceptors: "+", ".join(self.A),
         "Transitions:\n"+"\n".join(f" {q_i},{a},{z1}/{''.join(r[0])},{z2}/{''.join(r[1])} → {q_f}" for (q_i,a,z1,z2),(q_f,r) in self.items())]
        return "\n".join(info) 

    def print_summary(self):
        """ Prints a summary of the PDA
        """
        print(self.summary())

    def save_img(self,filename,q_c=set(),a_c=set(),q_prev=set(),symbols={},states={},format='svg',dpi="60.0",string=None,stack=[],status=None,one_arc=True,finished=False,cleanup=True):
        """ Saves machine as an image

        :param filename: Filename of image
        :param q_c: Set of current states to be highlited (default is empty)
        :param a_c: Set of current symbols to be highlited (default is empty)
        :param q_prev: Set of previos states to be highlited (default is empty)
        :param symbols: Replacements of the symbols to show (default is empty)
        :param states: Replacements of the states to show (default is empty)
        :param format: Format of image (default is svg)
        :param dpi: Resolution of image (default is "90.0")
        :param string: Label of string being analysed (default is None)
        :param stack: Status of the stack (default is None)
        :param status: Status of the PDA (default is None)
        :param one_arc: Graph one arc in case of multiple transitions (default is True)
        :param finished: If has pass through final state (default is False)

        :returns: None"""
        dot=self.graph(q_c=q_c,a_c=a_c,q_prev=q_prev,symbols=symbols,states=states,format=format,dpi=dpi,string=string,finished=finished,one_arc=one_arc,status=status,stack=stack)
        dot.render(filename,format=format,cleanup=cleanup)

    def graph(self,q_c=set(),a_c=set(),q_prev=set(),symbols={},states={},format="svg",dpi="60.0",string=None,stack=[],status=None,one_arc=True,finished=False):
        """ Graphs TSPDA

        :param q_c: Set of current states to be highlited (default is empty)
        :param a_c: Set of current symbols to be highlited (default is empty)
        :param q_prev: Set of previos states to be highlited (default is empty)
        :param symbols: Replacements of the symbols to show (default is empty)
        :param states: Replacements of the states to show (default is empty)
        :param format: Format of image (default is svg)
        :param dpi: Resolution of image (default is "90.0")
        :param string: Label of string being analysed (default is None)
        :param stack: Status of the stack (default is None)
        :param status: Status of the TSPDA (default is None)
        :param one_arc: Graph one arc in case of multiple transitions (default is True)
        :param finished: If has pass through final state (default is False)

        :returns: Returns Digraph object from graphviz"""
        if len(q_c)==0:
            states_=self._index2label(self.create_initial_istate())
        else:
            states_=q_c

        f=Digraph(comment="PDAs",format=format)
        f.attr(rankdir='LR',dpi=dpi)
        for i,((q_c_,stack),zetas) in enumerate(states_.items()):
            with f.subgraph(name=f'cluster_{i}') as f_:
                self._graph(f_,
                    i=i,
                    q_c=q_c_,
                    a_c=a_c,
                    q_prev=q_prev,
                    symbols=symbols,
                    states=states,
                    dpi=dpi,
                    format=format,
                    stack=stack,
                    zetas=zetas,
                    finished=finished,
                    status=status,
                    string=string,
                    one_arc=one_arc)
        return f

    def _graph(self,f,i=0,q_c=set(),a_c=set(),q_prev=set(),states={},symbols={},format="svg",dpi="60.0",string=None,stack=[],status=None,one_arc=True,zetas=[],finished=False):
        label_string=None
        label_stack=None
        if status==None:
            color_state="lightblue2"
        elif status and len(self.A.intersection(q_c))>0:
            color_state="limegreen"
        else:
            color_state="orangered"

        if string:
            l,c,r=string
            label_string=f"<TR><TD>{l}</TD> <TD><B>{c}</B></TD> <TD>{r}</TD></TR>"
            if finished and len(r)==0:
                if self.acceptor(q_c):
                    color_state="limegreen"
                else:
                    color_state="orangered"
        if stack:
            label_middle="".join([symbols.get(r_,r_) for r_ in stack[0]])
            label_stack1=f"<TR><TD ALIGN='RIGHT'>{label_middle}</TD></TR>"
            label_middle="".join([symbols.get(r_,r_) for r_ in stack[1]])
            label_stack2=f"<TR><TD ALIGN='RIGHT'>{label_middle}</TD></TR>"

        f.attr(style='invis',labelloc="b")
        if label_string and label_stack1:
            f.attr(label=f"< <TABLE BORDER='0' ><TR><TD><TABLE BORDER='1' CELLBORDER='0' SIDES='TBR'>{label_stack1}</TABLE></TD></TR><TR><TD><TABLE BORDER='1' CELLBORDER='0' SIDES='TBR'>{label_stack2}</TABLE></TD></TR><TR><TD><TABLE BORDER='0'>{label_string}</TABLE></TD></TR></TABLE>>")
        elif label_stack:
            f.attr(label=f"< <TABLE BORDER='1' CELLBORDER='0' SIDES='TBR'>{label_stack}</TABLE>>")

        for q,_ in self.Q.map.items():
            if q in self.A:
                shape="doublecircle"
            else:
                shape="circle"
            if q in q_c:
                f.node(name=f'{q}_{i}',label=states.get(q,q),shape=shape,color=color_state,style="filled")
            else:
                f.node(name=f'{q}_{i}',label=states.get(q,q),shape=shape)

        edges=defaultdict(list)
        for e,info in enumerate(self.items()):
            (q_i,a,z1,z2),(q_f,(r1,r2)) = info
            r1_=[symbols.get(r_,r_) for r_ in r1]
            r2_=[symbols.get(r_,r_) for r_ in r2]
            if (q_f in q_c and q_i in q_prev) and (a in a_c or a=="ε") and (z1,z2) in zetas:
                edges[(f'{q_i}_{i}',f'{q_f}_{i}')].append(
                        (f'{symbols.get(a,a)},{symbols.get(z1,z1)}/{"".join(r1_)},{symbols.get(z2,z2)}/{"".join(r2_)}',True))
            else:
                edges[(f'{q_i}_{i}',f'{q_f}_{i}')].append(
                        (f'{symbols.get(a,a)},{symbols.get(z1,z1)}/{"".join(r1_)},{symbols.get(z2,z2)}/{"".join(r2_)}',False))

        for (q_i,q_f),labels in edges.items():
            if one_arc:
                tags=[]
                colored_=False
                for label,colored in labels:
                    if colored:
                        colored_=True
                        tags.append(f'<FONT color="{color_state}">{label}</FONT>')
                    else:
                        tags.append(f'{label}')
                tags=f'< {"<BR/>".join(tags)} >'
                if colored_:
                    f.edge(q_i,q_f,label=tags,labelloc='b',color=color_state)
                else:
                    f.edge(q_i,q_f,label=tags,labelloc='b')
            elif not one_arc:
                for label,colored in labels:
                    if colored:
                        f.edge(q_i,q_f,label=label,labelloc='b',color=color_state,fontcolor=color_state)
                    else:
                        f.edge(q_i,q_f,label=label,labelloc='b')
        return f

    def table(self,symbols={},states={},q_order=None,s_order=None,color_final="#32a852",empty_symbol="∅"):
        """ Creates an HTML object for the table of the PDA

        :param symbols: Replacements of the symbols to show (default is empty)
        :param states: Replacements of the states to show (default is empty)
        :param  q_order: Order to use for states
        :param  s_order: Order to use for symbols
        :param color_final: RGB string for color of final state (default is "#32a852")

        :returns: Display object for IPython"""
        if not s_order:
            s_order=list(self.sigma)
            s_order.sort()
        if not q_order:
            q_order=list(self.Q)
            q_order.sort()
        symbs_h="</strong></td><td><strong>".join([symbols.get(q,q) for q in s_order])
        table=f"<table><tr><td></td><td><strong>{symbs_h}</strong></td></tr>"
        for q_i in q_order:
            vals=[]
            initial="⟶" if q_i == self.q_0 else ""
            final=f'bgcolor="{color_final}"' if q_i in self.A else ""
            vals.append(f"<strong>{initial}{states.get(q_i,q_i)}</strong>")
            for a in s_order:
                try:
                    labels=[]
                    for z1,info_ in self.ttable[self._nstate(q_i)][self._nsymbol(a)].items():
                        z1=self._qsymbol(z1)
                        for z2,info in info_.items():
                            z2=self._qsymbol(z2)
                            for q_f,(r1,r2) in info:
                                r1_=[symbols.get(self._qsymbol(r_),self._qsymbol(r_)) for r_ in r1]
                                r2_=[symbols.get(self._qsymbol(r_),self._qsymbol(r_)) for r_ in r2]
                                labels.append(
                                        f'{symbols.get(a,a)},{symbols.get(z1,z1)}/{"".join(r1_)},{symbols.get(z2,z2)}/{"".join(r2_)}'
                                        )
                    vals.append("<br/>".join(labels))
                except KeyError:
                    vals.append(empty_symbol)
            row="</td><td>".join(vals)
            table+=f"<tr><td {final}>{row}</td></tr>"
        table+="</table>"
        return display(HTML(table))

    def save_gif(self,w,filename="tspda.gif",symbols={},states={},dpi="90",show=True,loop=0,duration=500):
        """ Saves an animation of  machine

        :param w: String to analysed during animation
        :param filename: Name of gif (default is "tspda.gif")
        :param symbols: Replacements of the symbols to show (default is empty)
        :param states: Replacements of the states to show (default is empty)
        :param dpi: Resolution of image (default is "90.0")
        :param show: In interactive mode return gif
        :param loop: Number of loops in annimation, cero is forever (default is 0)
        :param duration: Duration in msegs among steps (default is 500)
        :returns: None or HTML for Ipython"""
        dirpath = tempfile.mkdtemp()
        i=0
        images=[]
        q_prev=set()
        max_images_height=1
        for q,a,w_,finished in self.delta_stepwise(w,mark_finished=True,index=True):
            filename_=os.path.join(dirpath,f'{i}')
            fin=len(w)-len(w_)
            processed=w[:max(0,int(i/2)-1)]
            a = a if a else " "
            g=self.save_img(filename_,q_c=q,a_c=set([a]),q_prev=q_prev,finished=finished,
                    symbols=symbols,states=states,
                    dpi=dpi,string=(processed,a,w_),format="png")
            q_prev=set([q_c for q_c,_ in q])
            im=Image.open(filename_+".png")
            width, height = im.size
            max_images_height=max(max_images_height,height)
            images.append(im)
            i+=1
            filename_=os.path.join(dirpath,f'{i}')
            g=self.save_img(filename_,q_c=q,
                    symbols=symbols,dpi=dpi,string=(processed,a,w_),finished=finished,
                    format="png")
            im=Image.open(filename_+".png")
            if i==0 or finished:
                images.append(im)
                images.append(im)
                images.append(im)
                images.append(im)
            images.append(im)
            i+=1
        for i,im in enumerate(images):
            im2 = Image.new('RGB', (width, max_images_height), (255, 255, 255))
            width, height = im.size
            im2.paste(im)
            images[i]=im2

        images[0].save(filename,
                save_all=True, append_images=images[1:], optimize=False, duration=duration, loop=loop)
        if show:
            return HTML(f'<img src="{filename}">')
        else:
            return




import types

class Language():
    def __init__(self,elements=[],max=30,finite=True,gen=None):
        """ Create a language 

        :param elements: This can be a finite enumeration of elements, another Language or a generator for an infinite language
        :param max: Maximum number of elements to print for infinite languages (default 30)
        :param finite: For languages based of generator if the language is finite or infinite, be aware theres is not check on this property and could raise errors (default True)"""
        self.max=max
        if isinstance(elements,Language):
            self.elements=elements.elements
            self.gen=elements.gen
            self.finite=elements.finite
        elif isinstance(gen,tuple):
            self.gen=gen
            self.elements=None
            self.finite=finite
        else:
            self.elements=set([str(e) for e in elements])
            self.gen=None
            self.finite=True

    def __iter__(self):
        if self.finite:
            for ele in self.elements:
                yield ele
        else:
            f,*args=self.gen
            yield from f(*args)

    def _epsilon(self,x):
        return 'ε' if len(x)==0 else str(x)

    def __repr__(self):
        if self.finite:
            return f"{{{','.join([self._epsilon(x) for x in self])}}}"
        else:
            return f"{{{','.join([self._epsilon(x) for _, x in zip(range(self.max), self)])},…}}"

    def __len__(self):
        if self.finite:
            return len(self.elements)
        else:
            return float('inf')

    def union(self,L):
        """ Union of the language with another language, if one is infintie returns an infinite generator

        :param l: Second Languauge fo the union
        :returns: Language  _self ∪ L_ """
        if self.finite and L.finite:
            return Language(self.elements.union(L),finite=True)
        else:
            return Language(gen=(self._union,L),finite=False)

    def _union(self,L):
        seen=set()
        for w in zip(self.elements,L):
            if not w in seen:
                seen.add(w)
                yield w
        for w in L:
            if not w in seen:
                seen.add(w)
                yield w

    def power(self,n):
        """ Power of a language 

        :returns: Language  _self^n_ """
        if n==0:
            return Language([""],finite=True)
        elif n==1:
            return Language(self)
        else:
            if self.finite:
                return Language(set(p for p in self._power(self,1,n)),finite=True)
            else:
                return Language(gen=(self._power,self,1,n),finite=False)

    def _power(self,partial,i,j):
        if i==j:
            for ele in partial:
                yield ele
        elif i<j:
            yield from self._power(self.concat(partial),i+1,j)


    def star(self):
        """ Star clouse for a language returns and infinite language 

        :returns: Language  _self*_ """
        return Language(gen=(self._star,),finite=False)

    def _star(self):
        yield ''
        prev=[""]
        seen=set([""])
        while True:
            for pre in prev:
                for ele in self:
                    new=pre+ele
                    if not new in seen:
                        prev.append(new)
                        seen.add(new)
                        yield new
            prev=[]

    def plus(self):
        """ Star clouse for a language returns an infinite language 

        :returns: Language  _self+_ """
        return Language(gen=(self._plus,),finite=False)

    def _plus(self):
        prev=[]
        seen=set()
        for e in self.elements:
            prev.append(e)
            yield  e
        while True:
            for pre in prev:
                for ele in self:
                    new=pre+ele
                    prev.append(new)
                    if not new in seen:
                        seen.add(new)
                        yield new
            prev=[]

    def concat(self,L):
        """ Concatenation of the language with another language, if one is infintie returns an infinite generator

        :param L: Second Languauge fo the concatenation
        :returns: Language  _self L_ """
        if self.finite and L.finite:
            return Language([w1+w2 for w2 in L for w1 in self],finite=True)
        else:
            return Language(gen=(self._concat,L),finite=False)

    def _concat(self,L):
        seen=set()
        for w2 in L:
            for w1 in self:
                new=w1+w2
                if not new in seen:
                    seen.add(new)
                    yield w1+w2


class Alphabet(set):
    """ Wrapper for the set python class"""

    def star(self):
        """ Star clouse for an alphabet returns an infinite language 

        :returns: Language  _self*_ """
        return Language(gen=(self._star,),finite=False)

    def _star(self):
        seen=set()
        yield ''
        prev=[""]
        while True:
            for pre in prev:
                for ele in self:
                    new=pre+ele
                    prev.append(new)
                    if not new in seen:
                        seen.add(new)
                        yield new
            prev=[]

    def power(self,n):
        """ Power of an alphabet

        :returns: Language  _Σ^n_ """
        if n==0:
            return Language([""],finite=True)
        elif n==1:
            return Language(self)
        else:
            return Language(set(p for p in self._power(self,1,n)),finite=True)

    def _power(self,partial,i,j):
        if i==j:
            for ele in partial:
                yield ele
        elif i<j:
            yield from self._power(Language([w1+w2 for w2 in partial for w1 in self],finite=True)
,i+1,j)





# Código para Autómata Finito [Determinístic]

from maquinas.exceptions import *
from maquinas.regular.ndfa_e import * 
from maquinas.regular.REParser import REParser

parser=REParser()

class RegularExpression():
    """Class for Regular Expressions

    :param er_string: regular expressiones.
    """
    def __init__(self, er_string):
        if len(er_string.strip())==0:
            raise NoStringWithDefinition(er_string,"RE")
        self.original=er_string
        self.ast=parser.parse(er_string,rulne_name="start")

    def __str__(self):
        return self.original

    def _ndfa_e(self, ast):
        if isinstance(ast,list):
            if len(ast)==3:
                fst=self._ndfa_e(ast[0])
                snd=self._ndfa_e(ast[2])
                fa=fst.union(snd)
            elif len(ast)==2:
                if not ast[1]=='*':
                    fst=self._ndfa_e(ast[0])
                    snd=self._ndfa_e(ast[1])
                    fa=fst.concat(snd)
                else:
                    fst=self._ndfa_e(ast[0])
                    fa=fst.kleene()
        else:
            if ast.startswith('"'):
                name=ast[1:-1]
            else:
                name=ast
            if name in ['empty','∅']:
                fa=ndfa_e_empty()
            elif name in ['epsilon','ε']:
                fa=ndfa_e_epsilon()
            else:
                fa=ndfa_e_single_symbol(name)
        return fa

    def ndfa_e(self):
        """ Generates a NDFA-e equivalent to the ER

        :returns: NDFA-e equivalent to ER"""
        return self._ndfa_e(self.ast)



# Código para Autómata Finito No Determinístico con transición epsilon

from maquinas.regular.fa import FiniteAutomaton
from maquinas.exceptions import *

from IPython.core.display import display, HTML
from graphviz import Digraph

epsilon='ε'

## TODO allow no strict
## TODO loaders
## TODO compress links in one
class NonDeterministicFiniteAutomaton_epsilon(FiniteAutomaton):
    """Class for Deterministic Finite Automaton

    :param Q: Ordered set of states (default is empty).
    :param sigma: Ordered set of symbols (default is empty).
    :param q_0: Initial state (default None).
    :param A: Set of acceptor states (default is empty).
    :param delta: List of transitions with the form tupla of tupla of q_i and a, and q_f (default is empty).
    :param force: If True and states or symbols do not exists create them (default is False).
    :type force: bool
    """
    def __init__(self, Q=[], sigma=[], q_0=None, A=[], delta=[], force=False):
        super().__init__(Q,sigma,q_0,A,delta,epsilon=True)
        self.e='ε'

    def expansion_epsilon(self,qs):
        """ Applies expansion by epsilon in a set of states

        :param qs: Set of states

        :returns: Set of reachable states from qs"""
        qs__=set(q for q in qs)
        change=True
        while change:
            qs_=set()
            for q in qs__:
                qs_.add(q)
                try:
                    qs_.update(self.get_transition(q,self.e))
                except DoesNotExistsTransition:
                    pass
            if len(qs__ ^ qs_)==0:
                change=False
            qs__=qs_
        return qs__

    def delta(self,q,a):
        """ Applies delta function

        :param q: Source state
        :param a: Symbol

        :returns: Destination state"""
        try:
            return self.expansion_epsilon(self.get_transition(q,a))
        except DoesNotExistsTransition:
            return set()

    def delta_extended(self,q,w):
        """ Applies delta extended function

        :param q: Source state
        :param w: String

        :returns: Destination state after processing the full string"""
        if q is None:
            q=set([self.q_0])
        if len(w)==0:
            return self.expansion_epsilon(q)
        else:
            *w_,a=w
            q_u=set()
            for r in self.delta_extended(q,w_):
                q_u.update(self.delta(r,a))
            return self.expansion_epsilon(q_u)

    def delta_stepwise(self,w,q=None):
        """ Applies a step of delta extended function

        :param w: String
        :param q: Source state (default is initial state)

        :returns: Tuple with state of precessing at step, consisting of: destination state, procesed symbol and processed string"""
        q_c= set([self.q_0]) if q is None else set(q)
        yield q_c,"",w
        for ix,a in enumerate(w):
            q_c=self.delta(q_c,a)
            q_c=self.expansion_epsilon(q_c)
            yield q_c,a,w[ix+1:]

    def graph(self,q_c=set(),a_c=set(),q_prev=set(),symbols={},states={},format="svg",dpi="60.0",string=None,**args):
        """ Graphs NDFA_e

        :param q_c: Set of current states to be highlited (default is empty)
        :param a_c: Set of current symbols to be highlited (default is empty)
        :param q_prev: Set of previos states to be highlited (default is empty)
        :param symbols: Replacements of the symbols to show (default is empty)
        :param states: Replacements of the states to show (default is empty)
        :param format: Format of image (default is svg)
        :param dpi: Resolution of image (default is "60.0")
        :param string: Label of string being analysed (default is None)
        :returns: Returns Digraph object from graphviz"""
        f=Digraph(comment="af",format=format)
        f.attr(rankdir='LR',dpi=dpi)
        color_state="lightblue2"
        if string:
            l,c,r=string
            f.attr(label=f"<{l} <B>{c}</B> {r}>")
            if len(r)==0:
                if len(self.A.intersection(q_c))>0:
                    color_state="limegreen"
                else:
                    color_state="orangered"
        f.node(name="__initial__",label="",shape="none",height=".0",width=".0")
        for q,q_ in self.Q.map.items():
            if q in self.A:
                shape="doublecircle"
            else:
                shape="circle"
            if q in q_c:
                f.node(name=q,label=states.get(q,q),shape=shape,color=color_state,style="filled")
            else:
                f.node(name=q,label=states.get(q,q),shape=shape)

        f.edge("__initial__",self.q_0)
        for e,info in enumerate(self.items()):
            (q_i,a),q_fs = info
            for q_f in q_fs:
                if (q_i in q_prev and a in a_c) or (q_prev and q_i in q_c and a == 'ε'):
                    f.edge(q_i,q_f,label=symbols.get(a,a),color=color_state,fontcolor=color_state)
                else:
                    f.edge(q_i,q_f,label=symbols.get(a,a))
        return f

    def table(self,symbols={},states={},q_order=None,s_order=None,color_final="#32a852",empty_symbol="∅"):
        """ Creates an HTML object for the table of the DFA

        :param symbols: Replacements of the symbols to show (default is empty)
        :param states: Replacements of the states to show (default is empty)
        :param  q_order: Order to use for states
        :param  s_order: Order to use for symbols
        :param color_final: RGB string for color of final state (default is "#32a852")

        :returns: Display object for IPython"""
        if not s_order:
            s_order=list(self.sigma)
            s_order.sort()
        if not q_order:
            q_order=list(self.Q)
            q_order.sort()
        symbs_h="</strong></td><td><strong>".join([symbols.get(q,q) for q in s_order])
        table=f"<table><tr><td></td><td><strong>{symbs_h}</strong></td></tr>"
        for q_i in q_order:
            vals=[]
            initial="⟶" if q_i == self.q_0 else "" 
            final=f'bgcolor="{color_final}"' if q_i in self.A else ""
            vals.append(f"<strong>{initial}{states.get(q_i,q_i)}</strong>")
            for a in s_order:
                try:
                    q_fs=set(states.get(q_f,q_f) for q_f in self[q_i,a])
                    if len(q_fs)>0:
                        vals.append(",".join(q_fs))
                    else:
                        vals.append(empty_symbol)
                except DoesNotExistsTransition:
                    vals.append(empty_symbol)
            row="</td><td>".join(vals)
            table+=f"<tr><td {final}>{row}</td></tr>"
        table+="</table>"
        return display(HTML(table))

    def union(self,other,label_a="ᴬ",label_b="ᴮ"):
        """Unites this NDFA_e with other NDFA-e

        :param other: Other NDFA-e
        :param label_a: Label to append to states of this NDFA-e (default is "ᴬ")
        :param label_b: Label to append to states of other NDFA-e (default is "ᴮ")

        :returns: Returns NDFA resultinf of the union of this an another NDFA-e"""
        new=NonDeterministicFiniteAutomaton_epsilon()
        new.add_state('q_0',initial=True)
        new.sigma=self.sigma.union(other.sigma)
        new.A=set()
        for q in self.Q:
            name=f"{q}{label_a}"
            new.Q.add(name)
            if q in self.A:
                new.A.add(name)
        for q in other.Q:
            name=f"{q}{label_b}"
            new.Q.add(name)
            if q in other.A:
                new.A.add(name)

        skip=1
        for nq_i,t_ in self.ttable.items():
            for na,nqs_f in t_.items():
                a=self.sigma.items[na]
                na=new.sigma.index(a)
                try:
                    new.ttable[nq_i+skip][na]=set([nq+skip for nq in nqs_f])
                except KeyError:
                    new.ttable[nq_i+skip]={}
                    new.ttable[nq_i+skip][na]=set([nq+skip for nq in nqs_f])

        skip=len(self.Q)+1
        for nq_i,t_ in other.ttable.items():
            for na,nqs_f in t_.items():
                a=other.sigma.items[na]
                na=new.sigma.index(a)
                try:
                    new.ttable[nq_i+skip][na]=set([nq+skip for nq in nqs_f])
                except KeyError:
                    new.ttable[nq_i+skip]={}
                    new.ttable[nq_i+skip][na]=set([nq+skip for nq in nqs_f])
        new.ttable[0]={}
        new.ttable[0][0]=set([1,skip])
        return new

    def concat(self,other,label_a="ᴬ",label_b="ᴮ",):
        """ Concatenates of this NDFA_e with other NDFA-e

        :param other: Other NDFA-e
        :param label_a: Label to append to states of this NDFA-e (default is "ᴬ")
        :param label_b: Label to append to states of other NDFA-e (default is "ᴮ")

        :returns: Returns NDFA resulting of the concatenation of this an another NDFA-e"""
        new=NonDeterministicFiniteAutomaton_epsilon()
        new.sigma=self.sigma.union(other.sigma)
        new.A=set()
        for q in self.Q:
            name=f"{q}{label_a}"
            if q==self.q_0:
                new.set_initial_state(name,force=True)
            new.Q.add(name)
        for q in other.Q:
            name=f"{q}{label_b}"
            new.Q.add(name)
            if q in other.A:
                new.A.add(name)

        skip=0
        for nq_i,t_ in self.ttable.items():
            for na,nqs_f in t_.items():
                a=self.sigma.items[na]
                na=new.sigma.index(a)
                try:
                    new.ttable[nq_i+skip][na]=set([nq+skip for nq in nqs_f])
                except KeyError:
                    new.ttable[nq_i+skip]={}
                    new.ttable[nq_i+skip][na]=set([nq+skip for nq in nqs_f])

        skip=len(self.Q)
        for nq_i,t_ in other.ttable.items():
            for na,nqs_f in t_.items():
                a=other.sigma.items[na]
                na=new.sigma.index(a)
                try:
                    new.ttable[nq_i+skip][na]=set([nq+skip for nq in nqs_f])
                except KeyError:
                    new.ttable[nq_i+skip]={}
                    new.ttable[nq_i+skip][na]=set([nq+skip for nq in nqs_f])

        for q in self.A:
            nq=self.Q.index(q)
            if nq in new.ttable:
                if 0 in new.ttable[nq]:
                    new.ttable[nq][0].add(skip)
                else:
                    new.ttable[nq][0]=set([skip])
            else:
                new.ttable[nq]={}
                new.ttable[nq][0]=set([skip])
        return new

    def kleene(self,label="ᴬ"):
        """Applies Kleene closure to this NDFA_e 

        :param label: Label to append to states of this NDFA-e (default is "ᴬ")

        :returns: Returns NDFA resulting of the application of Kleen closure to this NDFA-e"""
        new=NonDeterministicFiniteAutomaton_epsilon()
        new.add_state('q_0',initial=True)
        new.sigma=self.sigma 
        new.A=set()
        for q in self.Q:
            name=f"{q}{label}"
            new.Q.add(name)
        new.add_state('q_f')
        new.A.add('q_f')
        nq_f=len(new.Q)-1

        skip=1
        for nq_i,t_ in self.ttable.items():
            for na,nqs_f in t_.items():
                a=self.sigma.items[na]
                na=new.sigma.index(a)
                try:
                    new.ttable[nq_i+skip][na]=set([nq+skip for nq in nqs_f])
                except KeyError:
                    new.ttable[nq_i+skip]={}
                    new.ttable[nq_i+skip][na]=set([nq+skip for nq in nqs_f])
        new.ttable[0]={}
        new.ttable[0][0]=set([1,nq_f])
        new.ttable[nq_f]={}
        new.ttable[nq_f][0]=set([0])

        for q in self.A:
            nq=self.Q.index(q)
            if nq+skip in new.ttable:
                if 0 in new.ttable[nq+skip]:
                    new.ttable[nq+skip][0].add(nq_f)
                else:
                    new.ttable[nq+skip][0]=set([nq_f])
            else:
                new.ttable[nq+skip]={}
                new.ttable[nq+skip][0]=set([nq_f])

        return new


def ndfa_e_single_symbol(symbol):
    """NDFA-e for language of a single symbol

    :param symbol: Symbol to be accepted by NDFA-e

    :returns: NDFA-e for language of a single symbol"""
    fa=NonDeterministicFiniteAutomaton_epsilon()
    fa.add_transition("q_0",symbol,["q_1"],force=True)
    fa.set_initial_state("q_0")
    fa.set_aceptors(['q_1'])
    return fa

def ndfa_e_epsilon():
    """NDFA-e for empty string language

    :returns: NDFA-e for empty string language"""
    fa=NonDeterministicFiniteAutomaton_epsilon()
    fa.add_transition("q_0",epsilon,["q_0"],force=True)
    fa.set_initial_state("q_0",force=True)
    fa.set_aceptors(['q_0'])
    return fa

def ndfa_e_empty():
    """NDFA-e for empty language

    :returns: NDFA-e for empty language"""
    fa=NonDeterministicFiniteAutomaton_epsilon()
    fa.set_initial_state("q_0",force=True)
    fa.set_aceptors([])
    return fa



from ordered_set import OrderedSet
from collections import defaultdict, OrderedDict, Counter
from IPython.display import HTML, display
from graphviz import Graph
from copy import deepcopy
from maquinas.exceptions import NoSymbolInMachine

class PartialProduction():
    def __init__(self,head,body,lps=None):
        self.head=head
        # TODO: check for errors in tuplñe
        if isinstance(body,tuple):
            self.l=tuple(body[0])
            self.r=tuple(body[1])
        if not lps:
            self.lps=OrderedDict()
        else:
            self.lps=lps

    def __key(self):
        return (self.head, self.l, self.r)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, PartialProduction):
            return self.__key() == other.__key()
        return False

    def set_pointer(self,p1,p2,label):
        if p2:
            p=(p1,p2)
        else:
            p=(p1,)
        try:
            self.lps[label].add(p)
        except KeyError:
            self.lps[label]=OrderedSet([p])

    def is_partial_production(self):
        return not len(self.r)==0

    def get_next_symbol(self):
        return self.r[0]

    def get_head(self):
        return self.head

    def move_next_symbol(self):
        return PartialProduction(self.head,
                (self.l+self.r[:1],self.r[1:]))

    def move_prev_symbol(self):
        return PartialProduction(self.head,
                (self.l[:-1],self.l[-1:]+self.r))

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if len(self.l)>0 or len(self.r)>0:
            l="".join(self.l)
            r="".join(self.r)
            return f'{self.head} → {l}●{r}'
        else:
            return f'{self.head}'

    def __format__(self,format):
        return self.__str__()

    def pointers2str(self):
        res=[]
        for l,ps in self.lps.items():
            l=f"{l}"
            ps=", ".join([f'{p}' for p in ps ])
            res.append(f"{l}:{ps}")
        return "; ".join(res)

    def get_pointers(self,l):
        try:
            return self.lps[l]
        except IndexError:
            return []
        except KeyError:
            return []

    def reduce_name(self):
        if len(self.r)>0:
            l="".join(self.l)
            r="".join(self.r)
            return f'{self.head} → {l}●{r}'
        else:
            return f'{self.head}'

    def is_terminal(self):
        return len(self.l)==0 and len(self.r)==0

    def is_final(self,root):
        return self.head == root and len(self.r)==0

class Item():
    def __init__(self,pp,ini,fin=None):
        self.pp=pp
        self.ini=ini
        self.fin=fin

    def get_ini(self):
        return self.ini

    def get_fin(self):
        return self.fin

    def get_pproduction(self):
        return self.pp

    def get_head(self):
        return self.pp.get_head()

    def get_next_symbol(self):
        return self.pp.get_next_symbol()

    def get_span(self):
        if self.fin:
            return self.ini,self.fin
        else:
            return None

    def is_partial_production(self):
        return self.pp.is_partial_production()

    def get(self):
        return self.pp, self.ini, self.fin

    def reduce_name(self):
        if not self.fin is None:
            return f'{self.pp.reduce_name()} {self.ini}∷{self.fin}'
        else:
            return f'{self.pp.reduce_name()} {self.ini}'

    def get_pointers(self,label):
        return self.pp.get_pointers(label)

    def add_pointers(self,pointers):
        return self.pp.add_pointers(pointers)

    def pointers2str(self):
        return self.pp.pointers2str()

    def __str__(self):
        if not self.fin is None:
            return f'{self.pp} {self.ini}∷{self.fin}'
        else:
            return f'{self.pp} {self.ini}'

    def __repr__(self):
        return self.__str__()

    def __format__(self,format):
        return self.__str__()

    def __key(self):
        return (self.pp, self.ini, self.fin)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, Item):
            return self.__key() == other.__key()
        return False

    def move_next_symbol(self,ini,fin=None):
        return Item(
                self.pp.move_next_symbol(),
                ini=ini,fin=fin)

    def index(self):
        return Item(self.pp,self.ini)

    def is_final(self,root):
        return self.pp.is_final(root) and self.ini==0

    def is_terminal(self):
        return self.pp.is_terminal()

    def set_pointer(self,p1,p2,l):
        self.pp.set_pointer(p1,p2,l)

# Predictor function
def predictor(item,k,productions,working_table):
    pp=item.get_pproduction()
    next_symbol=pp.get_next_symbol()
    for body in productions[next_symbol]:
        pp=PartialProduction(next_symbol,([],body))
        working_table.add(Item(pp,k))

# Scanner function
def scanner(item,k,tokens,working_table,chart,terminals):
    symbol=item.get_next_symbol()
    origin=item.get_ini()
    token=tokens[k] if k<len(tokens) else None
    if symbol==token and symbol in terminals:
        new_item=item.move_next_symbol(origin)
        if new_item in chart[k+1]:
            ix=chart[k+1].index(new_item)
            new_item=chart[k+1][ix]
        item_=Item(item.pp,origin,k)
        if len(item.pp.l)>0:
            new_item.set_pointer(item_,None,k)
        chart[k+1].add(new_item)
    elif symbol=='ε':
        new_item=item.move_next_symbol(origin)
        if new_item in chart[k]:
            ix=chart[k].index(new_item)
            new_item=chart[k][ix]
        item_=Item(item.pp,origin,k)
        working_table.add(new_item)
        if len(item.pp.l)>0:
            new_item.set_pointer(item_,None,k)

def completer(item,k,working_table,chart):
    origin = item.get_ini()
    head= item.get_head()
    for prev_item in chart[origin]:
        try:
            symbol_canditate=prev_item.get_next_symbol()
        except IndexError:
            continue
        if symbol_canditate==head:
            new_item=prev_item.move_next_symbol(prev_item.ini)
            if new_item in chart[k]:
                ix=chart[k].index(new_item)
                new_item=chart[k][ix]
            elif new_item in working_table:
                ix=working_table.index(new_item)
                new_item=working_table[ix]
            item_=Item(item.pp,origin,k)
            prev_item_=Item(prev_item.pp,prev_item.ini,origin)
            if len(prev_item.pp.l)>0:
                new_item.set_pointer(item_,prev_item_,origin)
            else:
                new_item.set_pointer(item_,None,origin)
            working_table.add(new_item)

def parse(tokens,productions,root,terminals):
    chart = [ OrderedSet() for _ in range(len(tokens)+1)]
    for rule in productions[root]:
        starting_production=PartialProduction(root,([],rule))
        chart[0].add(Item(starting_production,0))
    pointers=defaultdict(set)
    for k in range(len(tokens)+1):
        working_table = OrderedSet(chart[k])
        chart[k].clear()
        # TODO: record the operation
        while len(working_table)>0:
            item=working_table.pop()
            if item in chart[k]:
                if (len(item.pp.r)>0 and item.pp.r[0]=='ε') or (len(item.pp.l)>0 and item.pp.l[0]=='ε'):
                    pass
                else:
                    ix=chart[k].index(item)
                    item_=chart[k][ix]
                    for l,ps in item.pp.lps.items():
                        item_.pp.lps[l].update(ps)
                    continue
            else:
                chart[k].add(item)

            if item.is_partial_production():
                symbol=item.get_next_symbol()
                if not symbol in terminals and not symbol=='ε':
                    predictor(item,k,productions,working_table)
                else:
                    scanner(item,k,tokens,working_table,chart,terminals)
            else:
                completer(item,k,working_table,chart)
    return chart

def print_chart(string,chart,pointers=False,lspace=40):
    for i,items in enumerate(chart):
        print(f"S({i}):","{}•{}".format(string[0:i],string[i:]))
        pointers_str=""
        for j,item in enumerate(items):
            pp=str(item.pp)
            if pointers:
                pointers_str=item.pointers2str()
            print(f"{j:3d} | {pp:20s} |  {item.ini:3d} | {pointers_str}")

def chart2table(string,chart,pointers=False):
    if pointers:
        tbl=["<table><tr><th>No.</th><th>Production</th><th>Origin</th><th>Pointers</th></tr>"]
    else:
        tbl=["<table><tr><th>No.</th><th>Production</th><th>Origin</th></tr>"]
    for i,items in enumerate(chart):
        tbl.append('<tr><th style="text-align:center" colspan="3">')
        tbl.append(f"S({i}):")
        tbl.append(f"{string[:i]}•{string[i:]}")
        tbl.append("</th></tr>")
        for j,item in enumerate(items):
            pp=str(item.pp)
            tbl.append("<tr>")
            tbl.append(f"<td>{j:3d}</td>")
            tbl.append(f"<td style='text-align:left'>{pp}</td>")
            tbl.append(f"<td>{item.ini:3d}</td>")
            if pointers:
                pointers_str=item.pointers2str()
                tbl.append(f"<td>{pointers_str}</td>")
            tbl.append("</td>")
    tbl.append("</table>")
    return display(HTML("".join(tbl)))

def item2label(item,span=False,symbols={}):
    label=f"{item[0]}"
    label=symbols.get(label,label)
    if span:
        label=f"<{label} <FONT POINT-SIZE='7'>{item[1]}:{item[2]}</FONT>>"
    return label


def _extract_trees(stack,nodes_dict,terminals,ancesters=Counter(),max_depth=None,max_ancesters=None,tokens=0):
    nchars=0
    depth=0
    id=1
    while len(stack)>0:
        root,current,children,partials,nchars,depth,ancesters=stack.pop(0)
        if max_depth and depth >= max_depth:
            #yield root
            continue
        prev,currentlist=current
        if prev:
            ancesters[prev['label']]+=1
        if len(children)==0 :
            if (prev['label'][0] in terminals or prev['label'][0]=="ε") and sum([len(p) for p in partials])==0:
                yield root
            else:
                _update_stack(stack,root,partials,nchars,terminals,prev=prev,depth=depth,ancesters=ancesters)
        elif len(children)==1:
            c=nodes_dict[children[0]]
            u=({'label':c['label'],
                'id':c['id'],
                'partial':c['partial'],
                },[])
            id+=1
            if max_ancesters and max_ancesters<=ancesters[c['label']]:
                continue
            currentlist.append(u)
            partials[0].insert(0,(u,c['children']))
            _update_stack(stack,root,partials,nchars,terminals,prev=prev,depth=depth,ancesters=ancesters)
        elif len(children)==2:
            c1=nodes_dict[children[0]]
            u1=({'label':c1['label'],
                'id':c1['id'],
                'partial':c1['partial'],
                },[])
            id+=1
            c2=nodes_dict[children[1]]
            u2=({'label':c2['label'],
                'id':c2['id'],
                'partial':c2['partial'],
                },[])
            id+=1
            if max_ancesters and ( max_ancesters<=ancesters[c1['label']] or max_ancesters<=ancesters[c2['label']]):
                continue
            currentlist.append(u1)
            currentlist.append(u2)
            partials[0].insert(0,(u1,c1['children']))
            partials[0].insert(1,(u2,c2['children']))
            _update_stack(stack,root,partials,nchars,terminals,prev=prev,depth=depth,ancesters=ancesters)
    return stack

def _deepcopy(tree,label,link=None):
    p,children=tree
    if 'current' in p and p['current']:
        link=(p,[c for c in children])
        return link,link
    children_=[]
    for c in children:
        c_,link_=_deepcopy(c,label,link=link)
        children_.append(c_)
        if link_:
            link=link_
    return (p,children_),link

def _update_stack(stack,root,partials,nchars,terminals,prev=None,depth=0,ancesters=Counter()):
    if len(partials)==0:
        if prev:
            if prev['label'][1]==prev['label'][2]:
                stack.insert(0,(root,(None,[]),[],[],nchars,depth,ancesters))
            else:
                stack.insert(0,(root,(None,[]),[],[],nchars+1,depth+1,ancesters))
        return
    while len(partials)>0 and len(partials[0])==0:
        partials.pop(0)
    if len(partials)==0:
        return
    current,branches=partials[0].pop(0)
    current[0]['current']=True
    if len(branches)==0:
        if current[0]['label'][0] in terminals :
            stack.insert(0,(root,current,[],partials,nchars+1,depth+1,ancesters))
        elif current[0]['label'][0]=='ε':
            stack.insert(0,(root,current,[],partials,nchars,depth+1,ancesters))
        else:
            raise NoSymbolInMachine(current[0]['label'][0],terminals)
    if len(branches)>0:
        root_,current_=_deepcopy(root,current[0])
        if current_ is None:
            partials.append((current,branches))
            current[0]['current']=False
        else:
            stack.append((root,current,branches[0],partials,nchars,depth+1,ancesters))
            for ix in range(1,len(branches)):
                root_,current_=_deepcopy(root,current[0])
                partials_=[deepcopy(p) for p in partials]
                stack.append((root_,current_,branches[ix],partials_,int(nchars),depth+1,Counter(ancesters)))
    current[0]['current']=False

def extract_trees(forest,terminals,max_depth=None,max_ancesters=None):
    u0,forest=forest
    u0_={'label':u0['label'],
        'id':u0['id'],
        'partial':u0['partial'],
        }
    root=(u0_,[])
    _,currentlist=root
    stack=[]
    _update_stack(stack,root,[[(root,u0['children'])]],0,terminals,ancesters=Counter())
    return _extract_trees(stack,forest,terminals,max_depth=max_depth,max_ancesters=max_ancesters,tokens=u0['tokens'])

def _graph_tree(f,tree,parent,
        c=0,
        ntree=0,
        symbols={},
        span=False,
        full=False,
        show_id=False,
        **params):
    tree=list(tree)
    if len(tree)>1:
        u,children=tree
    else:
        u,=tree
        children=()
    label=item2label(u['label'],span=span,symbols=symbols)
    u_n=f"u_{c}_{ntree}"
    if show_id:
        params['xlabel']=f"<<I><FONT POINT-SIZE='8'>u</FONT><FONT POINT-SIZE='5'>{u['id']+1}</FONT></I>>"
    if not u['partial'] or full:
        f.node(u_n,
            label=label,**params)
        if parent:
            f.edge(parent,u_n)
    else:
        u_n=parent

    for child in children:
        c+=1;
        c=_graph_tree(f,child,u_n,c=c,
                ntree=ntree,
                symbols=symbols,
                span=span,
                full=full,
                show_id=show_id,
                **params)
    return c

def graph_tree(tree,
        symbols={},
        span=False,
        full=False,
        show_id=False,
        node_params={'style':"rounded",
            'shape':"plain"},
        **params):
    params['format']='svg'
    params['strict']=True
    f=Graph(comment="tree",**params)
    c=0
    _graph_tree(f,tree,None,c=c,
                    symbols=symbols,
                    span=span,
                    full=full,
                    show_id=show_id,
                    **node_params)
    return f

def graph_trees(trees,
        symbols={},
        span=False,
        full=False,
        show_id=False,
        node_params={'style':"rounded",
            'shape':"plain"},
        **params):
    params['format']='svg'
    params['strict']=True
    f=Graph(comment="tree",**params)
    for i,tree in enumerate(trees):
        with f.subgraph(name=f'Tree {i}') as f_:
            _graph_tree(f_,tree,None,
                    ntree=i,
                    symbols=symbols,
                    span=span,
                    full=full,
                    show_id=show_id,
                    **node_params)
    return f

def _graph_forest(f,u,forest,parent,ancesters,processed,
        symbols={},
        span=False,
        full=True,
        show_id=False,
        **params):
    if not u:
        return

    n=tuple(forest.keys()).index(u['label'])
    u_n=f"u_{u['id']+1}"
    ancesters.add(u['id'])

    node_name=u_n
    original_node_name=u_n
    label=item2label(u['label'],span=span,symbols=symbols)

    if show_id:
        params['xlabel']=f"<<I><FONT POINT-SIZE='8'>u</FONT><FONT POINT-SIZE='5'>{u['id']+1}</FONT></I>>"
    if not u['partial'] or full:
        f.node(u_n,
                label=label,**params)
        if parent:
            if not (parent,original_node_name) in processed:
                f.edge(parent,original_node_name,dir="forward")
                processed.add((parent,original_node_name))
    else:
        node_name=parent

    for j,u_child_names in enumerate(u['children']):
            if len(u['children'])>1:
                fake_parent=f'fp_{j}_{original_node_name}'
                f.node(fake_parent,label="",shape='point')
                if not (original_node_name,fake_parent) in processed:
                    f.edge(original_node_name,fake_parent)
                    processed.add((original_node_name,fake_parent))
                node_name=fake_parent
            for u_child_name in u_child_names:
                u_child=forest[u_child_name]
                if u_child['id'] in ancesters:
                    u_n_child=f"u_{u_child['id']+1}"
                    if not (node_name,u_n_child) in processed:
                        f.edge(node_name,u_n_child,dir="forward")
                        processed.add((node_name,u_n_child))
                    continue
                else:
                    _graph_forest(f,u_child,forest,node_name,set(ancesters),processed,
                                symbols=symbols,
                                span=span,
                                full=full,
                                show_id=show_id,
                                **params)

def graph_forest(forest,
        symbols={},
        span=False,
        full=True,
        show_id=False,
        node_params={'style':"rounded",
            'shape':"rectangle"},
        **params):
    params['format']='svg'
    #params['strict']=True
    f=Graph(comment="tree",**params)
    u0,forest=forest
    _graph_forest(f,u0,forest,None,set(),set(),
                    symbols=symbols,
                    span=span,
                    full=full,
                    show_id=show_id,
                    **node_params)
    return f

def _make_node(label,elements):
    partial = False if isinstance(label[0],str) else True
    return {'label':label,
            'id':len(elements),
            'partial': partial,
            'children':OrderedSet()} #families

def _extract_forest(u,item,elements,processed,chart,terminals,i_table):
    if len(item.pp.l)==0:
        return None 

    symbol=item.pp.l[-1]
    pos=len(item.pp.l[:-1])
    first=len(item.pp.l)==1

    if symbol == 'ε':
        v_name=(item.pp.head,i_table,i_table)
        v=_make_node(v_name,elements)
        if not v_name in elements:
            elements[v_name]=v
        e_name=('ε',i_table,i_table)
        e=_make_node(e_name,elements)
        if not e_name in elements:
            elements[e_name]=e
        u['children'].add((e_name,))
        processed.add(item)
        return

    if symbol in terminals and first:
        v_name=(symbol,i_table-1,i_table)
        v=_make_node(v_name,elements)
        if not v_name in elements:
            elements[v_name]=v
        u['children'].add((v_name,))

    if (not symbol in terminals) and first:
        j=item.ini
        v_name=(symbol,j,i_table)
        v=_make_node(v_name,elements)
        if not v_name in elements:
            elements[v_name]=v
        u['children'].add((v_name,))
        for q, in item.pp.lps[j]:
            if not q in processed:
                _extract_forest(v,q,elements,processed,chart,terminals,i_table=i_table)

    if symbol in terminals and not first:
        j=item.ini
        v_name=(symbol,i_table-1,i_table)
        v=_make_node(v_name,elements)
        if not v_name in elements:
            elements[v_name]=v
        w_name=(item.pp.move_prev_symbol(),j,i_table-1)
        w=_make_node(w_name,elements)
        if not w_name in elements:
            elements[w_name]=w
            for p_, in item.pp.lps[i_table-1]:
                _extract_forest(w,p_,elements,processed,chart,terminals,i_table=i_table-1)
        u['children'].add((w_name,v_name))


    if (not symbol in terminals) and not first:
        j=item.ini
        for l,qps in item.pp.lps.items():
            for ii,(q,p_) in enumerate(qps):
                w_name=(item.pp.move_prev_symbol(),j,l)
                w=_make_node(w_name,elements)
                if not w_name in elements:
                    elements[w_name]=w
                else:
                    w=elements[w_name]
                if not p_ in processed:
                    processed.add(p_)
                    _extract_forest(w,p_,elements,processed,chart,terminals,i_table=l)
                v_name=(symbol,l,i_table)
                v=_make_node(v_name,elements)
                if not v_name in elements:
                    elements[v_name]=v
                else:
                    v=elements[v_name]
                if not q in processed:
                    processed.add(q)
                    _extract_forest(v,q,elements,processed,chart,terminals,i_table=i_table)
                u['children'].add((w_name,v_name))


def extract_forest(root,chart,terminals,tokens):
    elements=OrderedDict()
    processed=OrderedSet()
    i_table=len(chart)-1
    root_name=(root,0,i_table)
    u0=_make_node(root_name,elements)
    elements[u0['label']]=u0
    u0['tokens']=tokens
    for item in chart[-1]:
        if item.is_final(root):
            _extract_forest(u0,item,elements,processed,chart,terminals,i_table)
    return u0,elements

def verify_chart(chart,tokens,root):
    if not (len(tokens)+1) == len(chart):
        return False
    return [Item(item.pp,item.ini,len(tokens)) for item in chart[-1] if item.is_final(root)]



# Function to load machines
import re
from maquinas.regular.dfa import * 
from maquinas.regular.ndfa import * 
from maquinas.regular.ndfa_e import * 
from maquinas.exceptions import * 

re_state = re.compile(r"^\s*(?P<initial>->)?\s*(?P<state>.*[^\]])[^\]]*(?P<final>])?$")

def load_fa(string):
    header=False
    delta={}
    for line in string.split("\n"):
        line=line.strip()
        if len(line)==0 or line.startswith("#"):
            continue
        if not header:
            sigma=read_sigma(line)
            header=True
        else:
            origin,states=read_row(line)
            delta[origin]=states
    type_machine=1
    if 'epsilon' in sigma:
        type_machine=3
    if 'ε' in sigma:
        type_machine=3
    if type_machine==1:
        for o,states in delta.items():
            for s in states:
                if len(s)>1:
                    type_machine=2
                    break
            if type_machine==2:
                break
    if type_machine==1:
        m=DeterministicFiniteAutomaton()
    elif type_machine==2:
        m=NonDeterministicFiniteAutomaton()
    elif type_machine==3:
        m=NonDeterministicFiniteAutomaton_epsilon()

    for a in sigma:
        try:
            m.add_symbol(a)
        except AlreadyExistsSymbol:
            pass
    A=set()
    for (ini,fin,q_i),states in delta.items():
        if fin:
            A.add(q_i)
        for a,state in zip(sigma,states):
            state=[s for s in state if len(s)>0]
            if len(state)>0:
                m.add_transition(q_i,a,state,force=True)
        if ini:
            m.set_initial_state(q_i)
    m.set_aceptors(A)
    return m

def read_sigma(line):
    return [ a.strip() for a in  line.split("|") if len(a.strip())>0 ]

def read_row(line):
    row=[ a.strip() for a in  line.split("|") ]
    origin=read_state(row[0])
    return origin, read_states(row[1:])

def read_state(state):
    m=re_state.match(state)
    return m.group('initial')!=None,m.group('final')!=None,m.group('state').strip()

def read_states(states):
    return [[s.strip() for s in  state.split(',')] for state in states ]




# Código para Estado

from maquinas.exceptions import *
from maquinas.recursivelyenumerable.tm import TuringMachine

class Status():
    def __init__(self, machine, string="", state = None, memory=None):
        self.step=0
        self.status=None
        self.position=0
        self.machine=machine
        self.string=string
        if not machine is None:
            self.state = self.machine.delta_extended(None,"")
        else:
            self.state = None
        self.set_tape(string)
        self.memory=memory
        self.history=[]
        self.positioncolor="#b3e0ff"
        self.sucesscolor="##00b300"
        self.failcolor="#ff6666"

    def __str__(self):
        r= { "step": self.step,
             "tape": [_ for _ in self.tape],
             "pos": self.position,
             "state": self.state}
        return str(r)

    def set_tape(self,string):
        if isinstance(self.machine,TuringMachine):
            self.tape=None
            for state in self.machine.delta_stepwise(string):
                self.state = state
                break
        else:
            self.tape=f"{string}"

    def get_symbol_tape(self):
        try:
            return self.tape[self.position]
        except IndexError:
            self.status= self.machine.accept(self.state)

    def get_memory(self):
        return self.memory

    def next(self,dir=-1):
        if not self.status is None:
            return None
        if self.step==len(self.history):
            prev= { "step": self.step,
                    "tape": [_ for _ in self.tape] if self.tape else None,
                    "pos": self.position,
                    "state": self.state}
            self.history.append(prev)
            try:
                self.machine.stepStatus(self)
            except DoesNotExistsSymbol:
                self.status = False
        else:
            step=self.history[self.step+1]
            self.step=step['step']
            self.tape=step['tape']
            self.position=step['pos']
            self.state=step['state']
        if self.tape and self.position==len(self.tape):
            self.status= self.machine.acceptor(self.state)
        if not self.tape:
            if len(self.state)==0:
                self.status = False
                self.state=self.history[-1]['state']
            else:
                if self.machine.acceptor(self.state):
                    self.status = True

    def is_finish(self):
        if self.tape:
            return self.position==len(self.tape)
        else:
            return False

    def tape2html(self,ncols=20,symbols={},states={}):
        if self.tape:
            tape_idx=list(enumerate(self.tape))
            tape=tape_idx[max(0,self.position-ncols):min(2*ncols,self.position+ncols+1)]
            columns=[]
            for pos,symbol in tape:
                symbol_=symbols.get(symbol,symbol)
                if self.status is None:
                    if pos==self.position:
                        columns.append(f"<td  bgcolor='{self.positioncolor}'>{symbol_}</td>")
                    else:
                        columns.append(f"<td>{symbol_}</td>")
                elif self.status:
                    columns.append(f"<td  bgcolor='{self.sucesscolor}'>{symbol_}</td>")
                else:
                    columns.append(f"<td  bgcolor='{self.failcolor}'>{symbol_}</td>")
            columns="".join(columns)
            row=f'<tr>{columns}</tr>'
        if self.status is None:
            status="Processing"
        elif self.status:
            status="Accepted"
        else:
            status="Rejected"

        state_=self.machine._status(self.state,states=states,symbols=symbols)
        if self.tape:
            table=f'<strong>Estado actual:</strong> {state_}<br/><strong>Cadena:</strong> {status}<table>{row}</table>'
        else:
            table=f'<strong>Estado actual:</strong> {state_}<br/><strong>'
        return table
